
# Slixmpp: The Slick XMPP Library
# Copyright (C) 2016 Emmanuel Gil Peyrot <linkmauve@linkmauve.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

import slixmpp
from slixmpp.stanza import Message
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin, ElementBase, ET
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0380 import stanza, Encryption


log = logging.getLogger(__name__)


class XEP_0380(BasePlugin):

    """
    XEP-0380: Explicit Message Encryption
    """

    name = 'xep_0380'
    description = 'XEP-0380: Explicit Message Encryption'
    dependencies = {'xep_0030'}
    default_config = {
        'template': 'This message is encrypted with {name} ({namespace})',
    }

    mechanisms = {
        'jabber:x:encrypted': 'Legacy OpenPGP',
        'urn:xmpp:ox:0': 'OpenPGP for XMPP',
        'urn:xmpp:otr:0': 'OTR',
        'eu.siacs.conversations.axolotl': 'Legacy OMEMO',
        'urn:xmpp:omemo:0': 'OMEMO',
    }

    def plugin_init(self):
        self.xmpp.register_handler(
            Callback('Explicit Message Encryption',
                     StanzaPath('message/eme'),
                     self._handle_eme))

        register_stanza_plugin(Message, Encryption)

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(Encryption.namespace)

    def has_eme(self, msg):
        return msg.xml.find('{%s}encryption' % Encryption.namespace) is not None

    def add_eme(self, msg: Message, namespace: str) -> Message:
        msg['eme']['name'] = self.mechanisms[namespace]
        msg['eme']['namespace'] = namespace
        return msg

    def replace_body_with_eme(self, msg):
        eme = msg['eme']
        namespace = eme['namespace']
        name = self.mechanisms[namespace] if namespace in self.mechanisms else eme['name']
        body = self.config['template'].format(name=name, namespace=namespace)
        msg['body'] = body

    def _handle_eme(self, msg):
        self.xmpp.event('message_encryption', msg)



import unittest
from slixmpp import Presence
from slixmpp.test import SlixTest
import slixmpp.plugins.xep_0317 as xep_0317
from slixmpp.plugins.xep_0317 import stanza


class TestStanzaHats(SlixTest):

    def setUp(self):
        stanza.register_plugin()

    def test_create_hats(self):
        raw_xml = """
          <hats xmlns="urn:xmpp:hats:0">
            <hat uri="http://example.com/hats#Teacher" title="Teacher"/>
          </hats>
        """

        hats = xep_0317.Hats()

        hat = xep_0317.Hat()
        hat['uri'] = 'http://example.com/hats#Teacher'
        hat['title'] = 'Teacher'
        hats.append(hat)

        self.check(hats, raw_xml, use_values=False)

    def test_set_single_hat(self):
        presence = Presence()
        presence["hats"]["hat"]["uri"] = "test-uri"
        presence["hats"]["hat"]["title"] = "test-title"
        self.check(
            presence,  # language=XML
            """
            <presence>
              <hats xmlns='urn:xmpp:hats:0'>
                <hat uri='test-uri' title='test-title'/>
              </hats>
            </presence>
            """,
        )

    def test_set_multi_hat(self):
        presence = Presence()
        presence["hats"].add_hats([("uri1", "title1"), ("uri2", "title2")])
        self.check(
            presence,  # language=XML
            """
            <presence>
              <hats xmlns='urn:xmpp:hats:0'>
                <hat uri='uri1' title='title1'/>
                <hat uri='uri2' title='title2'/>
              </hats>
            </presence>
            """,
        )

    def test_get_hats(self):
        presence = Presence()
        presence["hats"].add_hats([("uri1", "title1"), ("uri2", "title2")])
        for i, hat in enumerate(presence["hats"]["hats"], start=1):
            self.assertEqual(hat["uri"], f"uri{i}")
            self.assertEqual(hat["title"], f"title{i}")


suite = unittest.TestLoader().loadTestsFromTestCase(TestStanzaHats)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import Message
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0004 import stanza
from slixmpp.plugins.xep_0004.stanza import Form, FormField, FieldOption


class XEP_0004(BasePlugin):

    """
    XEP-0004: Data Forms
    """

    name = 'xep_0004'
    description = 'XEP-0004: Data Forms'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_handler(
            Callback('Data Form',
                 StanzaPath('message/form'),
                 self.handle_form))

        register_stanza_plugin(FormField, FieldOption, iterable=True)
        register_stanza_plugin(Form, FormField, iterable=True)
        register_stanza_plugin(Message, Form)

    def plugin_end(self):
        self.xmpp.remove_handler('Data Form')
        self.xmpp['xep_0030'].del_feature(feature='jabber:x:data')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('jabber:x:data')

    def make_form(self, ftype='form', title='', instructions=''):
        f = Form()
        f['type'] = ftype
        f['title'] = title
        f['instructions'] = instructions
        return f

    def handle_form(self, message):
        self.xmpp.event("message_xform", message)

    def build_form(self, xml):
        return Form(xml=xml)



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp


class AdminCommands(slixmpp.ClientXMPP):

    """
    A simple Slixmpp bot that uses admin commands to
    add a new user to a server.
    """

    def __init__(self, jid, password, command):
        slixmpp.ClientXMPP.__init__(self, jid, password)

        self.command = command

        self.add_event_handler("session_start", self.start)

    async def start(self, event):
        """
        Process the session_start event.

        Typical actions for the session_start event are
        requesting the roster and broadcasting an initial
        presence stanza.

        Arguments:
            event -- An empty dictionary. The session_start
                     event does not provide any additional
                     data.
        """
        self.send_presence()
        await self.get_roster()

        def command_success(iq, session):
            print('Command completed')
            if iq['command']['form']:
                for var, field in iq['command']['form']['fields'].items():
                    print('%s: %s' % (var, field['value']))
            if iq['command']['notes']:
                print('Command Notes:')
                for note in iq['command']['notes']:
                    print('%s: %s' % note)
            self.disconnect()

        def command_error(iq, session):
            print('Error completing command')
            print('%s: %s' % (iq['error']['condition'],
                              iq['error']['text']))
            self['xep_0050'].terminate_command(session)
            self.disconnect()

        def process_form(iq, session):
            form = iq['command']['form']
            answers = {}
            for var, field in form['fields'].items():
                if var != 'FORM_TYPE':
                    if field['type'] == 'boolean':
                        answers[var] = input('%s (y/n): ' % field['label'])
                        if answers[var].lower() in ('1', 'true', 'y', 'yes'):
                            answers[var] = '1'
                        else:
                            answers[var] = '0'
                    else:
                        answers[var] = input('%s: ' % field['label'])
                else:
                    answers['FORM_TYPE'] = field['value']
            form['type'] = 'submit'
            form['values'] = answers

            session['next'] = command_success
            session['payload'] = form

            self['xep_0050'].complete_command(session)

        session = {'next': process_form,
                   'error': command_error}

        command = self.command.replace('-', '_')
        handler = getattr(self['xep_0133'], command, None)

        if handler:
            handler(session={
                'next': process_form,
                'error': command_error
            })
        else:
            print('Invalid command name: %s' % self.command)
            self.disconnect()


if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()

    # Output verbosity options.
    parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                        action="store_const", dest="loglevel",
                        const=logging.ERROR, default=logging.INFO)
    parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                        action="store_const", dest="loglevel",
                        const=logging.DEBUG, default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")
    parser.add_argument("-c", "--command", dest="command",
                        help="admin command to use")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")
    if args.command is None:
        args.command = input("Admin command: ")

    # Setup the CommandBot and register plugins. Note that while plugins may
    # have interdependencies, the order in which you register them does
    # not matter.
    xmpp = AdminCommands(args.jid, args.password, args.command)
    xmpp.register_plugin('xep_0133') # Service Administration

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_forever()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
from asyncio import Future
from datetime import datetime, timedelta
from typing import (
    Dict,
    Optional
)

from slixmpp.plugins import BasePlugin
from slixmpp import JID
from slixmpp.stanza import Iq
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import CoroutineCallback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0012 import stanza, LastActivity


log = logging.getLogger(__name__)


class XEP_0012(BasePlugin):

    """
    XEP-0012 Last Activity
    """

    name = 'xep_0012'
    description = 'XEP-0012: Last Activity'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, LastActivity)

        self._last_activities = {}

        self.xmpp.register_handler(
            CoroutineCallback('Last Activity',
                 StanzaPath('iq@type=get/last_activity'),
                 self._handle_get_last_activity))

        self.api.register(self._default_get_last_activity,
                'get_last_activity',
                default=True)
        self.api.register(self._default_set_last_activity,
                'set_last_activity',
                default=True)
        self.api.register(self._default_del_last_activity,
                'del_last_activity',
                default=True)

    def plugin_end(self):
        self.xmpp.remove_handler('Last Activity')
        self.xmpp['xep_0030'].del_feature(feature='jabber:iq:last')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('jabber:iq:last')

    def begin_idle(self, jid: Optional[JID] = None, status: Optional[str] = None) -> Future:
        """Reset the last activity for the given JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.

        :param status: Optional status.
        """
        return self.set_last_activity(jid, 0, status)

    def end_idle(self, jid: Optional[JID] = None) -> Future:
        """Remove the last activity of a JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.del_last_activity(jid)

    def start_uptime(self, status: Optional[str] = None) -> Future:
        """
        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.set_last_activity(None, 0, status)

    def set_last_activity(self, jid=None, seconds=None, status=None) -> Future:
        """Set last activity for a JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.api['set_last_activity'](jid, args={
            'seconds': seconds,
            'status': status
        })

    def del_last_activity(self, jid: JID) -> Future:
        """Remove the last activity of a JID.

        .. versionchanged:: 1.8.0
            This function now returns a Future.
        """
        return self.api['del_last_activity'](jid)

    def get_last_activity(self, jid: JID, local: bool = False,
                          ifrom: Optional[JID] = None, **iqkwargs) -> Future:
        """Get last activity for a specific JID.

        :param local: Fetch the value from the local cache.
        """
        if jid is not None and not isinstance(jid, JID):
            jid = JID(jid)

        if self.xmpp.is_component:
            if jid.domain == self.xmpp.boundjid.domain:
                local = True
        else:
            if str(jid) == str(self.xmpp.boundjid):
                local = True
        jid = jid.full

        if local or jid in (None, ''):
            log.debug("Looking up local last activity data for %s", jid)
            return self.api['get_last_activity'](jid, None, ifrom, None)

        iq = self.xmpp.make_iq_get(ito=jid, ifrom=ifrom)
        iq.enable('last_activity')
        return iq.send(**iqkwargs)

    async def _handle_get_last_activity(self, iq: Iq):
        log.debug("Received last activity query from " + \
                  "<%s> to <%s>.", iq['from'], iq['to'])
        reply = await self.api['get_last_activity'](iq['to'], None, iq['from'], iq)
        reply.send()

    # =================================================================
    # Default in-memory implementations for storing last activity data.
    # =================================================================

    def _default_set_last_activity(self, jid: JID, node: str, ifrom: JID, data: Dict):
        seconds = data.get('seconds', None)
        if seconds is None:
            seconds = 0

        status = data.get('status', None)
        if status is None:
            status = ''

        self._last_activities[jid] = {
            'seconds': datetime.now() - timedelta(seconds=seconds),
            'status': status}

    def _default_del_last_activity(self, jid: JID, node: str, ifrom: JID, data: Dict):
        if jid in self._last_activities:
            del self._last_activities[jid]

    def _default_get_last_activity(self, jid: JID, node: str, ifrom: JID, iq: Iq) -> Iq:
        if not isinstance(iq, Iq):
            reply = self.xmpp.Iq()
        else:
            reply = iq.reply()

        if jid not in self._last_activities:
            raise XMPPError('service-unavailable')

        bare = JID(jid).bare

        if bare != self.xmpp.boundjid.bare:
            if bare in self.xmpp.roster[jid]:
                sub = self.xmpp.roster[jid][bare]['subscription']
                if sub not in ('from', 'both'):
                    raise XMPPError('forbidden')

        td = datetime.now() - self._last_activities[jid]['seconds']
        seconds = td.seconds + td.days * 24 * 3600
        status = self._last_activities[jid]['status']

        reply['last_activity']['seconds'] = seconds
        reply['last_activity']['status'] = status

        return reply



import unittest
from slixmpp.jid import JID
from slixmpp.test import SlixTest


class TestStreamCorrect(SlixTest):

    def test_recv_correct(self):
        self.stream_start(mode='client', plugins=['xep_0308'])

        recv = []

        def recv_correct(msg):
            recv.append(msg)

        self.xmpp.add_event_handler('message_correction', recv_correct)

        self.recv("""
            <message from="example.com" to="toto@example">
                <replace xmlns="urn:xmpp:message-correct:0" id="tototo"/>
                <body>oucou</body>
            </message>
        """)
        received = recv[0]
        self.assertEqual(received['replace']['id'], "tototo")


    def test_send_correct(self):
        self.stream_start(mode='client', plugins=['xep_0308'])

        corrected = self.xmpp.plugin['xep_0308'].build_correction(
            id_to_replace="12345",
            mto=JID('toto@example.com'),
            mbody="I am replacing",
        )
        self.assertEqual(corrected['replace']['id'], '12345')
        self.assertEqual(corrected['to'], JID('toto@example.com'))
        self.assertEqual(corrected['body'], 'I am replacing')
        corrected['id'] = 'my id'

        corrected = self.xmpp.plugin['xep_0308'].correct_message(
            corrected,
            'This is new',
        )
        self.send("""
            <message type="chat" to="toto@example.com">
                <body>This is new</body>
                <replace xmlns="urn:xmpp:message-correct:0" id="my id" />
            </message>
        """)


suite = unittest.TestLoader().loadTestsFromTestCase(TestStreamCorrect)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.xmlstream import ElementBase, ET


class UserNick(ElementBase):

    """
    XEP-0172: User Nickname allows the addition of a <nick> element
    in several stanza types, including <message> and <presence> stanzas.

    The nickname contained in a <nick> should be the global, friendly or
    informal name chosen by the owner of a bare JID. The <nick> element
    may be included when establishing communications with new entities,
    such as normal XMPP users or MUC services.

    The nickname contained in a <nick> element will not necessarily be
    the same as the nickname used in a MUC.

    Example stanzas:

    ::

        <message to="user@example.com">
          <nick xmlns="http://jabber.org/nick/nick">The User</nick>
          <body>...</body>
        </message>

        <presence to="otheruser@example.com" type="subscribe">
          <nick xmlns="http://jabber.org/nick/nick">The User</nick>
        </presence>

    Stanza Interface:

    ::

        nick -- A global, friendly or informal name chosen by a user.

    """

    namespace = 'http://jabber.org/protocol/nick'
    name = 'nick'
    plugin_attrib = name
    interfaces = {'nick'}

    def set_nick(self, nick):
        """
        Add a <nick> element with the given nickname.

        Arguments:
            nick -- A human readable, informal name.
        """
        self.xml.text = nick

    def get_nick(self):
        """Return the nickname in the <nick> element."""
        return self.xml.text

    def del_nick(self):
        """Remove the <nick> element."""
        if self.parent is not None:
            self.parent().xml.remove(self.xml)



import logging
from datetime import date
from typing import Optional

from slixmpp import (
    JID,
    ComponentXMPP,
    register_stanza_plugin,
)
from slixmpp.plugins.base import BasePlugin

from . import stanza


class XEP_0292(BasePlugin):
    """
    vCard4 over XMPP

    Does not implement the IQ semantics that neither movim does gajim implement,
    cf https://xmpp.org/extensions/xep-0292.html#self-iq-retrieval and
    https://xmpp.org/extensions/xep-0292.html#self-iq-publication

    Does not implement the "empty pubsub event item" as a notification mechanism,
    that neither gajim nor movim implement
    https://xmpp.org/extensions/xep-0292.html#sect-idm45744791178720

    Relies on classic pubsub semantics instead.
    """
    xmpp: ComponentXMPP

    name = "xep_0292"
    description = "vCard4 Over XMPP"
    dependencies = {"xep_0163", "xep_0060", "xep_0030"}
    stanza = stanza

    def plugin_init(self):
        pubsub_stanza = self.xmpp["xep_0060"].stanza

        register_stanza_plugin(pubsub_stanza.Item, stanza.VCard4)
        register_stanza_plugin(pubsub_stanza.EventItem, stanza.VCard4)

        self.xmpp['xep_0060'].map_node_event(stanza.NS, 'vcard4')

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=stanza.NS)
        self.xmpp['xep_0163'].remove_interest(stanza.NS)

    def session_bind(self, jid):
        self.xmpp['xep_0163'].register_pep('vcard4', stanza.VCard4)

    def publish_vcard(
        self,
        full_name: Optional[str] = None,
        given: Optional[str] = None,
        surname: Optional[str] = None,
        birthday: Optional[date] = None,
        nickname: Optional[str] = None,
        phone: Optional[str] = None,
        note: Optional[str] = None,
        url: Optional[str] = None,
        email: Optional[str] = None,
        country: Optional[str] = None,
        locality: Optional[str] = None,
        impp: Optional[str] = None,
        **pubsubkwargs,
    ):
        """
        Publish a vcard using PEP
        """
        vcard = stanza.VCard4()

        if impp:
            vcard.add_impp(impp)

        if nickname:
            vcard.add_nickname(nickname)
        if full_name:
            vcard["full_name"] = full_name

        if given:
            vcard["given"] = given
        if surname:
            vcard["surname"] = surname
        if birthday:
            vcard["birthday"] = birthday

        if note:
            vcard.add_note(note)
        if url:
            vcard.add_url(url)
        if email:
            vcard.add_email(email)
        if phone:
            vcard.add_tel(phone)
        if country and locality:
            vcard.add_address(country, locality)
        elif country:
            vcard.add_address(country, locality)

        return self.xmpp["xep_0163"].publish(vcard, id="current", **pubsubkwargs)

    def retrieve_vcard(self, jid: JID, **pubsubkwargs):
        """
        Retrieve a vcard using PEP
        """
        return self.xmpp["xep_0060"].get_item(
            jid, stanza.VCard4.namespace, "current", **pubsubkwargs
        )


log = logging.getLogger(__name__)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 nicoco
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import unittest
from xml.etree import ElementTree as ET

from slixmpp import register_stanza_plugin, ElementBase
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0492 import stanza
from slixmpp.plugins.xep_0402 import stanza as b_stanza


class TestNotificationSetting(SlixTest):
    def setUp(self):
        b_stanza.register_plugin()
        stanza.register_plugin()

    def test_never(self):
        bookmark = b_stanza.Conference()
        bookmark["extensions"]["notify"].configure("never")
        self.check(
            bookmark,
            """
            <conference xmlns='urn:xmpp:bookmarks:1'>
              <extensions>
                <notify xmlns='urn:xmpp:notification-settings:0'>
                  <never />
                </notify>
              </extensions>
            </conference>
            """,
            use_values=False,
        )

    def test_always(self):
        bookmark = b_stanza.Conference()
        bookmark["extensions"]["notify"].configure("always")
        self.check(
            bookmark,
            """
            <conference xmlns='urn:xmpp:bookmarks:1'>
              <extensions>
                <notify xmlns='urn:xmpp:notification-settings:0'>
                  <always />
                </notify>
              </extensions>
            </conference>
            """,
            use_values=False,
        )

    def test_on_mention(self):
        bookmark = b_stanza.Conference()
        bookmark["extensions"]["notify"].configure("on-mention")
        self.check(
            bookmark,
            """
            <conference xmlns='urn:xmpp:bookmarks:1'>
              <extensions>
                <notify xmlns='urn:xmpp:notification-settings:0'>
                  <on-mention />
                </notify>
              </extensions>
            </conference>
            """,
            use_values=False,
        )

    def test_advanced(self):
        bookmark = b_stanza.Conference()
        bookmark["extensions"]["notify"].configure("never", client_type="pc")
        bookmark["extensions"]["notify"].configure("on-mention", client_type="mobile")

        register_stanza_plugin(stanza.Advanced, AdvancedExtension)
        bookmark["extensions"]["notify"]["advanced"].enable("cool")
        bookmark["extensions"]["notify"]["advanced"]["cool"]["attrib"] = "cool-attrib"
        bookmark["extensions"]["notify"]["advanced"]["cool"]["content"] = "cool-content"
        self.check(
            bookmark,
            """
            <conference xmlns='urn:xmpp:bookmarks:1'>
              <extensions>
                <notify xmlns='urn:xmpp:notification-settings:0'>
                  <never client-type="pc" />
                  <on-mention client-type="mobile" />
                  <advanced>
                    <cool xmlns="cool-ns" attrib="cool-attrib">cool-content</cool>
                  </advanced>
                </notify>
              </extensions>
            </conference>
            """,
            use_values=False,
        )

    def test_change_config(self):
        bookmark = b_stanza.Conference()
        bookmark["extensions"]["notify"].configure("never")
        bookmark["extensions"]["notify"].configure("never", client_type="pc")
        bookmark["extensions"]["notify"].configure("on-mention", client_type="mobile")

        self.check(
            bookmark,
            """
            <conference xmlns='urn:xmpp:bookmarks:1'>
              <extensions>
                <notify xmlns='urn:xmpp:notification-settings:0'>
                  <never />
                  <never client-type="pc" />
                  <on-mention client-type="mobile" />
                </notify>
              </extensions>
            </conference>
            """,
            use_values=False,
        )

        bookmark["extensions"]["notify"].configure("always")

        self.check(
            bookmark,
            """
            <conference xmlns='urn:xmpp:bookmarks:1'>
              <extensions>
                <notify xmlns='urn:xmpp:notification-settings:0'>
                  <always />
                  <never client-type="pc" />
                  <on-mention client-type="mobile" />
                </notify>
              </extensions>
            </conference>
            """,
            use_values=False,
        )

        bookmark["extensions"]["notify"].configure("always", "mobile")

        self.check(
            bookmark,
            """
            <conference xmlns='urn:xmpp:bookmarks:1'>
              <extensions>
                <notify xmlns='urn:xmpp:notification-settings:0'>
                  <always />
                  <never client-type="pc" />
                  <always client-type="mobile" />
                </notify>
              </extensions>
            </conference>
            """,
            use_values=False,
        )

    def test_get_config(self):
        bookmark = b_stanza.Conference()
        bookmark["extensions"]["notify"].configure("never")
        bookmark["extensions"]["notify"].configure("never", client_type="pc")
        bookmark["extensions"]["notify"].configure("on-mention", client_type="mobile")

        self.assertEqual(bookmark["extensions"]["notify"].get_config(), "never")
        self.assertEqual(bookmark["extensions"]["notify"].get_config("pc"), "never")
        self.assertEqual(
            bookmark["extensions"]["notify"].get_config("mobile"), "on-mention"
        )

    def test_get_config_2(self):
        notify = stanza.Notify()
        notify.configure("always")
        self.assertEqual(notify.get_config(), "always")

        notify = stanza.Notify(
            xml=ET.fromstring(
                "<notify xmlns='urn:xmpp:notification-settings:0'><always /></notify>"
            )
        )
        self.assertEqual(notify.get_config(), "always")

        notify = stanza.Notify(
            xml=ET.fromstring(
                "<notify xmlns='urn:xmpp:notification-settings:0'>"
                "<always client-type='pc'/>"
                "</notify>"
            )
        )
        self.assertIsNone(notify.get_config())
        self.assertEqual(notify.get_config("pc"), "always")

    def test_add_config(self):
        notify = stanza.Notify(
            xml=ET.fromstring(
                "<notify xmlns='urn:xmpp:notification-settings:0'>"
                "<always client-type='pc'/>"
                "</notify>"
            )
        )
        notify.configure("never")
        self.check(
            notify,
            """
            <notify xmlns='urn:xmpp:notification-settings:0'>
                <never/>
                <always client-type='pc'/>
            </notify>
            """,
            use_values=False
        )

    def test_update_config(self):
        notify = stanza.Notify(
            xml=ET.fromstring(
                "<notify xmlns='urn:xmpp:notification-settings:0'>"
                "<always />"
                "</notify>"
            )
        )
        notify.configure("never")
        self.check(
            notify,
            """
            <notify xmlns='urn:xmpp:notification-settings:0'>
                <never/>
            </notify>
            """,
            use_values=False
        )

        notify = stanza.Notify(
            xml=ET.fromstring(
                "<notify xmlns='urn:xmpp:notification-settings:0'>"
                "<always client-type='pc'/>"
                "</notify>"
            )
        )
        notify.configure("never", client_type="pc")
        self.check(
            notify,
            """
            <notify xmlns='urn:xmpp:notification-settings:0'>
                <never client-type='pc'/>
            </notify>
            """,
            use_values=False
        )

class AdvancedExtension(ElementBase):
    namespace = "cool-ns"
    name = "cool"
    plugin_attrib = name
    interfaces = {"attrib", "content"}

    def set_content(self, content: str):
        self.xml.text = content


suite = unittest.TestLoader().loadTestsFromTestCase(TestNotificationSetting)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Erik Reuterborg Larsson, Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0184.stanza import Request, Received
from slixmpp.plugins.xep_0184.receipt import XEP_0184


register_plugin(XEP_0184)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission
from slixmpp.plugins.base import register_plugin

from . import stanza
from .sims import XEP_0385

register_plugin(XEP_0385)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Error
from slixmpp.xmlstream import ElementBase, StanzaBase

class ClientStateIndication(ElementBase):
    name = 'csi'
    namespace = 'urn:xmpp:csi:0'
    plugin_attrib = name

class Active(StanzaBase):
    name = 'active'
    plugin_attrib = 'active'
    namespace = 'urn:xmpp:csi:0'

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()

class Inactive (StanzaBase):
    name = 'inactive'
    plugin_attrib = 'inactive'
    namespace = 'urn:xmpp:csi:0'

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.xml.tag = self.tag_name()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp.xmlstream import JID
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0060 import stanza


log = logging.getLogger(__name__)


class XEP_0060(BasePlugin):

    """
    XEP-0060 Publish Subscribe
    """

    name = 'xep_0060'
    description = 'XEP-0060: Publish-Subscribe'
    dependencies = {'xep_0030', 'xep_0004', 'xep_0082', 'xep_0131'}
    stanza = stanza

    def plugin_init(self):
        self.node_event_map = {}

        self.xmpp.register_handler(
                Callback('Pubsub Event: Items',
                    StanzaPath('message/pubsub_event/items'),
                    self._handle_event_items))
        self.xmpp.register_handler(
                Callback('Pubsub Event: Purge',
                    StanzaPath('message/pubsub_event/purge'),
                    self._handle_event_purge))
        self.xmpp.register_handler(
                Callback('Pubsub Event: Delete',
                    StanzaPath('message/pubsub_event/delete'),
                    self._handle_event_delete))
        self.xmpp.register_handler(
                Callback('Pubsub Event: Configuration',
                    StanzaPath('message/pubsub_event/configuration'),
                    self._handle_event_configuration))
        self.xmpp.register_handler(
                Callback('Pubsub Event: Subscription',
                    StanzaPath('message/pubsub_event/subscription'),
                    self._handle_event_subscription))

        self.xmpp['xep_0131'].supported_headers.add('SubID')

    def plugin_end(self):
        self.xmpp.remove_handler('Pubsub Event: Items')
        self.xmpp.remove_handler('Pubsub Event: Purge')
        self.xmpp.remove_handler('Pubsub Event: Delete')
        self.xmpp.remove_handler('Pubsub Event: Configuration')
        self.xmpp.remove_handler('Pubsub Event: Subscription')

    def _handle_event_items(self, msg):
        """Raise events for publish and retraction notifications."""
        node = msg['pubsub_event']['items']['node']

        multi = len(msg['pubsub_event']['items']) > 1
        values = {}
        if multi:
            values = msg.values
            del values['pubsub_event']

        for item in msg['pubsub_event']['items']:
            event_name = self.node_event_map.get(node, None)
            event_type = 'publish'
            if item.name == 'retract':
                event_type = 'retract'

            if multi:
                condensed = self.xmpp.Message()
                condensed.values = values
                condensed['pubsub_event']['items']['node'] = node
                condensed['pubsub_event']['items'].append(item)
                self.xmpp.event('pubsub_%s' % event_type, msg)
                if event_name:
                    self.xmpp.event('%s_%s' % (event_name, event_type),
                                    condensed)
            else:
                self.xmpp.event('pubsub_%s' % event_type, msg)
                if event_name:
                    self.xmpp.event('%s_%s' % (event_name, event_type), msg)

    def _handle_event_purge(self, msg):
        """Raise events for node purge notifications."""
        node = msg['pubsub_event']['purge']['node']
        event_name = self.node_event_map.get(node, None)

        self.xmpp.event('pubsub_purge', msg)
        if event_name:
            self.xmpp.event('%s_purge' % event_name, msg)

    def _handle_event_delete(self, msg):
        """Raise events for node deletion notifications."""
        node = msg['pubsub_event']['delete']['node']
        event_name = self.node_event_map.get(node, None)

        self.xmpp.event('pubsub_delete', msg)
        if event_name:
            self.xmpp.event('%s_delete' % event_name, msg)

    def _handle_event_configuration(self, msg):
        """Raise events for node configuration notifications."""
        node = msg['pubsub_event']['configuration']['node']
        event_name = self.node_event_map.get(node, None)

        self.xmpp.event('pubsub_config', msg)
        if event_name:
            self.xmpp.event('%s_config' % event_name, msg)

    def _handle_event_subscription(self, msg):
        """Raise events for node subscription notifications."""
        node = msg['pubsub_event']['subscription']['node']
        event_name = self.node_event_map.get(node, None)

        self.xmpp.event('pubsub_subscription', msg)
        if event_name:
            self.xmpp.event('%s_subscription' % event_name, msg)

    def map_node_event(self, node, event_name):
        """
        Map node names to events.

        When a pubsub event is received for the given node,
        raise the provided event.

        For example::

            map_node_event('http://jabber.org/protocol/tune',
                           'user_tune')

        will produce the events 'user_tune_publish' and 'user_tune_retract'
        when the respective notifications are received from the node
        'http://jabber.org/protocol/tune', among other events.

        :param node: The node name to map to an event.
        :param event_name: The name of the event to raise when a
                          notification from the given node is received.
        """
        self.node_event_map[node] = event_name

    def create_node(self, jid, node, config=None, ntype=None, ifrom=None,
                    timeout_callback=None, callback=None, timeout=None):
        """
        Create and configure a new pubsub node.

        A server MAY use a different name for the node than the one provided,
        so be sure to check the result stanza for a server assigned name.

        If no configuration form is provided, the node will be created using
        the server's default configuration. To get the default configuration
        use get_node_config().

        :param jid: The JID of the pubsub service.
        :param node: Optional name of the node to create. If no name is
                     provided, the server MAY generate a node ID for you.
                     The server can also assign a different name than the
                     one you provide; check the result stanza to see if
                     the server assigned a name.
        :param config: Optional XEP-0004 data form of configuration settings.
        :param ntype: The type of node to create. Servers typically default
                      to using 'leaf' if no type is provided.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')
        iq['pubsub']['create']['node'] = node

        if config is not None:
            form_type = 'http://jabber.org/protocol/pubsub#node_config'
            if 'FORM_TYPE' in config.get_fields():
                config.field['FORM_TYPE']['value'] = form_type
            else:
                config.add_field(var='FORM_TYPE',
                                 ftype='hidden',
                                 value=form_type)
            if ntype:
                if 'pubsub#node_type' in config.get_fields():
                    config.field['pubsub#node_type']['value'] = ntype
                else:
                    config.add_field(var='pubsub#node_type', value=ntype)
            iq['pubsub']['configure'].append(config)

        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def subscribe(self, jid, node, bare=True, subscribee=None, options=None,
                  ifrom=None, timeout_callback=None, callback=None,
                  timeout=None):
        """
        Subscribe to updates from a pubsub node.

        The rules for determining the JID that is subscribing to the node are:
        1. If subscribee is given, use that as provided.
        2. If ifrom was given, use the bare or full version based on bare.
        3. Otherwise, use self.xmpp.boundjid based on bare.

        :param jid: The pubsub service JID.
        :param node: The node to subscribe to.
        :param bare: Indicates if the subscribee is a bare or full JID.
                     Defaults to True for a bare JID.
        :param subscribee: The JID that is subscribing to the node.
        :param options:
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')
        iq['pubsub']['subscribe']['node'] = node

        if subscribee is None:
            if ifrom:
                if bare:
                    subscribee = JID(ifrom).bare
                else:
                    subscribee = ifrom
            else:
                if bare:
                    subscribee = self.xmpp.boundjid.bare
                else:
                    subscribee = self.xmpp.boundjid

        iq['pubsub']['subscribe']['jid'] = subscribee
        if options is not None:
            iq['pubsub']['options'].append(options)
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def unsubscribe(self, jid, node, subid=None, bare=True, subscribee=None,
                    ifrom=None, timeout_callback=None, callback=None,
                    timeout=None):
        """
        Unubscribe from updates from a pubsub node.

        The rules for determining the JID that is unsubscribing
        from the node are:
        1. If subscribee is given, use that as provided.
        2. If ifrom was given, use the bare or full version based on bare.
        3. Otherwise, use self.xmpp.boundjid based on bare.

        :param jid: The pubsub service JID.
        :param node: The node to unsubscribe from.
        :param subid: The specific subscription, if multiple subscriptions
                      exist for this JID/node combination.
        :param bare: Indicates if the subscribee is a bare or full JID.
                     Defaults to True for a bare JID.
        :param subscribee: The JID that is unsubscribing from the node.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')
        iq['pubsub']['unsubscribe']['node'] = node

        if subscribee is None:
            if ifrom:
                if bare:
                    subscribee = JID(ifrom).bare
                else:
                    subscribee = ifrom
            else:
                if bare:
                    subscribee = self.xmpp.boundjid.bare
                else:
                    subscribee = self.xmpp.boundjid

        iq['pubsub']['unsubscribe']['jid'] = subscribee
        iq['pubsub']['unsubscribe']['subid'] = subid
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def get_subscriptions(self, jid, node=None, ifrom=None,
                          timeout_callback=None, callback=None,
                          timeout=None):
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='get')
        iq['pubsub']['subscriptions']['node'] = node
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def get_affiliations(self, jid, node=None, ifrom=None,
                         timeout_callback=None, callback=None, timeout=None):
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='get')
        iq['pubsub']['affiliations']['node'] = node
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def get_subscription_options(self, jid, node=None, user_jid=None,
                                 ifrom=None, timeout_callback=None,
                                 callback=None, timeout=None):
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='get')
        if user_jid is None:
            iq['pubsub']['default']['node'] = node
        else:
            iq['pubsub']['options']['node'] = node
            iq['pubsub']['options']['jid'] = user_jid
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def set_subscription_options(self, jid, node, user_jid, options,
                                 ifrom=None, timeout_callback=None,
                                 callback=None, timeout=None):
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='get')
        iq['pubsub']['options']['node'] = node
        iq['pubsub']['options']['jid'] = user_jid
        iq['pubsub']['options'].append(options)
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def get_node_config(self, jid, node=None, ifrom=None,
                        timeout_callback=None, callback=None, timeout=None):
        """
        Retrieve the configuration for a node, or the pubsub service's
        default configuration for new nodes.

        :param jid: The JID of the pubsub service.
        :param node: The node to retrieve the configuration for. If None,
                     the default configuration for new nodes will be
                     requested. Defaults to None.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='get')
        if node is None:
            iq['pubsub_owner']['default']
        else:
            iq['pubsub_owner']['configure']['node'] = node
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def get_node_subscriptions(self, jid, node, ifrom=None,
                               timeout_callback=None, callback=None,
                               timeout=None):
        """
        Retrieve the subscriptions associated with a given node.

        :param jid: The JID of the pubsub service.
        :param node: The node to retrieve subscriptions from.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='get')
        iq['pubsub_owner']['subscriptions']['node'] = node
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def get_node_affiliations(self, jid, node, ifrom=None, timeout_callback=None,
                              callback=None, timeout=None):
        """
        Retrieve the affiliations associated with a given node.

        :param jid: The JID of the pubsub service.
        :param node: The node to retrieve affiliations from.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='get')
        iq['pubsub_owner']['affiliations']['node'] = node
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def delete_node(self, jid, node, ifrom=None, timeout_callback=None, callback=None,
                    timeout=None):
        """
        Delete a a pubsub node.

        :param jid: The JID of the pubsub service.
        :param node: The node to delete.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')
        iq['pubsub_owner']['delete']['node'] = node
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def set_node_config(self, jid, node, config, ifrom=None,
                        timeout_callback=None, callback=None, timeout=None):
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')
        iq['pubsub_owner']['configure']['node'] = node
        iq['pubsub_owner']['configure'].append(config)
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def publish(self, jid, node, id=None, payload=None, options=None,
                ifrom=None, timeout_callback=None, callback=None,
                timeout=None):
        """
        Add a new item to a node, or edit an existing item.

        For services that support it, you can use the publish command
        as an event signal by not including an ID or payload.

        When including a payload and you do not provide an ID then
        the service will generally create an ID for you.

        Publish options may be specified, and how those options
        are processed is left to the service, such as treating
        the options as preconditions that the node's settings
        must match.

        :param jid: The JID of the pubsub service.
        :param node: The node to publish the item to.
        :param id: Optionally specify the ID of the item.
        :param payload: The item content to publish.
        :param options: A form of publish options.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')
        iq['pubsub']['publish']['node'] = node
        if id is not None:
            iq['pubsub']['publish']['item']['id'] = id
        if payload is not None:
            iq['pubsub']['publish']['item']['payload'] = payload
        iq['pubsub']['publish_options'] = options
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def retract(self, jid, node, id, notify=None, ifrom=None,
                timeout_callback=None, callback=None, timeout=None):
        """
        Delete a single item from a node.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')

        iq['pubsub']['retract']['node'] = node
        iq['pubsub']['retract']['notify'] = notify
        iq['pubsub']['retract']['item']['id'] = id
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def purge(self, jid, node, ifrom=None, timeout_callback=None, callback=None,
              timeout=None):
        """
        Remove all items from a node.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')
        iq['pubsub_owner']['purge']['node'] = node
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def get_nodes(self, *args, **kwargs):
        """
        Discover the nodes provided by a Pubsub service, using disco.
        """
        return self.xmpp['xep_0030'].get_items(*args, **kwargs)

    def get_item(self, jid, node, item_id, ifrom=None,
                 timeout_callback=None, callback=None, timeout=None):
        """
        Retrieve the content of an individual item.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='get')
        item = stanza.Item()
        item['id'] = item_id
        iq['pubsub']['items']['node'] = node
        iq['pubsub']['items'].append(item)
        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def get_items(self, jid, node, item_ids=None, max_items=None,
                  iterator=False, ifrom=None, timeout_callback=None,
                  callback=None, timeout=None):
        """
        Request the contents of a node's items.

        The desired items can be specified, or a query for the last
        few published items can be used.

        Pubsub services may use result set management for nodes with
        many items, so an iterator can be returned if needed.
        """
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='get')
        iq['pubsub']['items']['node'] = node
        iq['pubsub']['items']['max_items'] = max_items

        if item_ids is not None:
            for item_id in item_ids:
                item = stanza.Item()
                item['id'] = item_id
                iq['pubsub']['items'].append(item)

        if iterator:
            return self.xmpp['xep_0059'].iterate(iq, 'pubsub')
        else:
            return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def get_item_ids(self, jid, node, ifrom=None, timeout_callback=None, callback=None,
                     timeout=None, iterator=False):
        """
        Retrieve the ItemIDs hosted by a given node, using disco.
        """
        return self.xmpp['xep_0030'].get_items(jid, node, ifrom=ifrom,
                                               callback=callback, timeout=timeout,
                                               iterator=iterator,
                                               timeout_callback=timeout_callback)

    def modify_affiliations(self, jid, node, affiliations=None, ifrom=None,
                            timeout_callback=None, callback=None,
                            timeout=None):
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')
        iq['pubsub_owner']['affiliations']['node'] = node

        if affiliations is None:
            affiliations = []

        for jid, affiliation in affiliations:
            aff = stanza.OwnerAffiliation()
            aff['jid'] = jid
            aff['affiliation'] = affiliation
            iq['pubsub_owner']['affiliations'].append(aff)

        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)

    def modify_subscriptions(self, jid, node, subscriptions=None,
                             ifrom=None, timeout_callback=None,
                             callback=None, timeout=None):
        iq = self.xmpp.Iq(sto=jid, sfrom=ifrom, stype='set')
        iq['pubsub_owner']['subscriptions']['node'] = node

        if subscriptions is None:
            subscriptions = []

        for jid, subscription in subscriptions:
            sub = stanza.OwnerSubscription()
            sub['jid'] = jid
            sub['subscription'] = subscription
            iq['pubsub_owner']['subscriptions'].append(sub)

        return iq.send(callback=callback, timeout=timeout, timeout_callback=timeout_callback)



# Slixmpp: The Slick XMPP Library
# Copyright © 2021 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

"""
This file contains boilerplate to define types relevant to slixmpp.
"""

from typing import (
    Any,
    Dict,
    Optional,
    Union,
    Iterable,
    List,
)

try:
    from typing import (
        Literal,
        TypedDict,
        Protocol,
    )
except ImportError:
    from typing_extensions import (
        Literal,
        TypedDict,
        Protocol,
    )

from slixmpp.jid import JID

PresenceTypes = Literal[
    'error', 'probe', 'subscribe', 'subscribed',
    'unavailable', 'unsubscribe', 'unsubscribed',
]

PresenceShows = Literal[
    'away', 'chat', 'dnd', 'xa',
]

# add the empty string, but not for sending
ExtPresenceShows = Literal[
    'away', 'chat', 'dnd', 'xa', ''
]

MessageTypes = Literal[
    'chat', 'error', 'groupchat',
    'headline', 'normal',
]

IqTypes = Literal[
    "error", "get", "set", "result",
]

MucRole = Literal[
    'moderator', 'participant', 'visitor', 'none'
]

MucAffiliation = Literal[
    'outcast', 'member', 'admin', 'owner', 'none'
]

OptJid = Optional[JID]
JidStr = Union[str, JID]
OptJidStr = Optional[Union[str, JID]]

class PresenceArgs(TypedDict, total=False):
    pfrom: JidStr
    pto: JidStr
    pshow: PresenceShows
    ptype: PresenceTypes
    pstatus: str


class MucRoomItem(TypedDict, total=False):
    jid: str
    role: MucRole
    affiliation: MucAffiliation
    show: Optional[PresenceShows]
    status: str
    alt_nick: str


class ResourceDict(TypedDict, total=False):
    show: ExtPresenceShows
    priority: int
    status: str


RosterState = TypedDict(
    'RosterState',
    {
        'from': bool,
        'to': bool,
        'pending_in': bool,
        'pending_out': bool,
        'whitelisted': bool,
        'subscription': str,
        'name': str,
        'groups': List[str],
        'removed': bool,
    }
)


class RosterDBProtocol(Protocol):
    def load(self, owner:JidStr, jid:JidStr, db_state: Dict[str, Any]) -> Optional[RosterState]:
        ...

    def save(self, owner:JidStr, jid:JidStr, state: RosterState, db_state: Dict[str, Any]):
        ...

    def entries(self, owner: OptJidStr, db_state: Optional[dict[str, Any]] = None) -> Iterable[str]:
        ...


MucRoomItemKeys = Literal[
    'jid', 'role', 'affiliation', 'show', 'status',  'alt_nick',
]

MAMDefault = Literal['always', 'never', 'roster']

FilterString = Literal['in', 'out', 'out_sync']

ErrorTypes = Literal["modify", "cancel", "auth", "wait", "cancel"]

ErrorConditions = Literal[
    "bad-request",
    "conflict",
    "feature-not-implemented",
    "forbidden",
    "gone",
    "internal-server-error",
    "item-not-found",
    "jid-malformed",
    "not-acceptable",
    "not-allowed",
    "not-authorized",
    "payment-required",
    "policy-violation",
    "recipient-unavailable",
    "redirect",
    "registration-required",
    "remote-server-not-found",
    "remote-server-timeout",
    "resource-constraint",
    "service-unavailable",
    "subscription-required",
    "undefined-condition",
    "unexpected-request",
]

# https://xmpp.org/registrar/disco-categories.html#client
ClientTypes = Literal[
    "bot",
    "console",
    "game",
    "handheld",
    "pc",
    "phone",
    "sms",
    "tablet",
    "web",
]

__all__ = [
    'Protocol', 'TypedDict', 'Literal', 'OptJid', 'OptJidStr', 'JidStr', 'MAMDefault',
    'PresenceTypes', 'PresenceShows', 'MessageTypes', 'IqTypes', 'MucRole',
    'MucAffiliation', 'FilterString', 'ErrorConditions', 'ErrorTypes', 'ClientTypes'
]




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0013.stanza import Offline
from slixmpp.plugins.xep_0013.offline import XEP_0013


register_plugin(XEP_0013)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
import logging

from slixmpp.stanza import Message, Error, StreamFeatures
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.matcher import StanzaPath, MatchMany
from slixmpp.xmlstream.handler import Callback
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0079 import stanza


log = logging.getLogger(__name__)


class XEP_0079(BasePlugin):

    """
    XEP-0079 Advanced Message Processing
    """

    name = 'xep_0079'
    description = 'XEP-0079: Advanced Message Processing'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, stanza.AMP)
        register_stanza_plugin(Error, stanza.InvalidRules)
        register_stanza_plugin(Error, stanza.UnsupportedConditions)
        register_stanza_plugin(Error, stanza.UnsupportedActions)
        register_stanza_plugin(Error, stanza.FailedRules)

        self.xmpp.register_handler(
                Callback('AMP Response',
                    MatchMany([
                        StanzaPath('message/error/failed_rules'),
                        StanzaPath('message/amp')
                    ]),
                    self._handle_amp_response))

        if not self.xmpp.is_component:
            self.xmpp.register_feature('amp',
                    self._handle_amp_feature,
                    restart=False,
                    order=9000)
            register_stanza_plugin(StreamFeatures, stanza.AMPFeature)

    def plugin_end(self):
        self.xmpp.remove_handler('AMP Response')

    def _handle_amp_response(self, msg):
        log.debug('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>')
        if msg['type'] == 'error':
            self.xmpp.event('amp_error', msg)
        elif msg['amp']['status'] in ('alert', 'notify'):
            self.xmpp.event('amp_%s' % msg['amp']['status'], msg)

    def _handle_amp_feature(self, features):
        log.debug('Advanced Message Processing is available.')
        self.xmpp.features.add('amp')

    def discover_support(self, jid=None, **iqargs):
        if jid is None:
            if self.xmpp.is_component:
                jid = self.xmpp.server_host
            else:
                jid = self.xmpp.boundjid.host

        return self.xmpp['xep_0030'].get_info(
                jid=jid,
                node='http://jabber.org/protocol/amp',
                **iqargs)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

import slixmpp
from slixmpp.plugins.xep_0030 import StaticDisco


log = logging.getLogger(__name__)


class StaticExtendedDisco(object):

    """
    Extend the default StaticDisco implementation to provide
    support for extended identity information.
    """

    def __init__(self, static):
        """
        Augment the default XEP-0030 static handler object.

        Arguments:
            static -- The default static XEP-0030 handler object.
        """
        self.static = static

    def set_extended_info(self, jid, node, ifrom, data):
        """
        Replace the extended identity data for a JID/node combination.

        The data parameter may provide:
            data -- Either a single data form, or a list of data forms.
        """
        self.del_extended_info(jid, node, ifrom, data)
        self.add_extended_info(jid, node, ifrom, data)

    def add_extended_info(self, jid, node, ifrom, data):
        """
        Add additional extended identity data for a JID/node combination.

        The data parameter may provide:
            data -- Either a single data form, or a list of data forms.
        """
        self.static.add_node(jid, node)

        forms = data.get('data', [])
        if not isinstance(forms, list):
            forms = [forms]

        info = self.static.get_node(jid, node)['info']
        for form in forms:
            info.append(form)

    def del_extended_info(self, jid, node, ifrom, data):
        """
        Replace the extended identity data for a JID/node combination.

        The data parameter is not used.
        """
        if self.static.node_exists(jid, node):
            info = self.static.get_node(jid, node)['info']
            for form in info['substanza']:
                info.xml.remove(form.xml)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
import logging
from typing import Optional

from slixmpp.stanza import Message
from slixmpp.jid import JID
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0308 import stanza, Replace


log = logging.getLogger(__name__)


class XEP_0308(BasePlugin):

    """
    XEP-0308 Last Message Correction
    """

    name = 'xep_0308'
    description = 'XEP-0308: Last Message Correction'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_handler(
            Callback('Message Correction',
                     StanzaPath('message/replace'),
                     self._handle_correction))

        register_stanza_plugin(Message, Replace)

        self.xmpp.use_message_ids = True

    def plugin_end(self):
        self.xmpp.remove_handler('Message Correction')
        self.xmpp.plugin['xep_0030'].del_feature(feature=Replace.namespace)

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(Replace.namespace)

    def is_correction(self, msg: Message):
        return msg.xml.find('{%s}replace' % Replace.namespace) is not None

    def _handle_correction(self, msg: Message):
        self.xmpp.event('message_correction', msg)

    def build_correction(self, id_to_replace: str, mto: JID,
                         mfrom: Optional[JID] = None, mtype: str = 'chat',
                         mbody: str = '') -> Message:
        """
        Build a corrected message.

        :param id_to_replace: The id of the original message.
        :param mto: Recipient of the message, must be the same as the original
                    message.
        :param mfrom: Sender of the message, must be the same as the original
                      message.
        :param mtype: Type of the message, must be the send as the original
                      message.
        :param mbody: The corrected message body.
        """
        msg = self.xmpp.make_message(
            mto=mto,
            mfrom=mfrom,
            mbody=mbody,
            mtype=mtype
        )
        msg['replace']['id'] = id_to_replace
        return msg

    def correct_message(self, msg: Message, body: str) -> Message:
        """
        Send a correction to an existing message.

        :param msg: The message that must be replaced.
        :param body: The body to set in the correcting message.
        :returns: The message that was sent.
        """
        to_replace = msg['id']
        mto = msg['to']
        mfrom = msg['from']
        mtype = msg['type']
        if not to_replace:
            raise ValueError('No available ID for replacing the message')
        if not mto:
            raise ValueError('No available recipient JID')

        new = self.build_correction(
            id_to_replace=to_replace,
            mto=mto,
            mfrom=mfrom,
            mtype=mtype,
            mbody=body,
        )
        new.send()
        return new




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import datetime as dt

from slixmpp.jid import JID
from slixmpp.xmlstream import ElementBase
from slixmpp.plugins import xep_0082


class LegacyDelay(ElementBase):

    name = 'x'
    namespace = 'jabber:x:delay'
    plugin_attrib = 'legacy_delay'
    interfaces = {'from', 'stamp', 'text'}

    def get_from(self):
        from_ = self._get_attr('from')
        return JID(from_) if from_ else None

    def set_from(self, value):
        self._set_attr('from', str(value))

    def get_stamp(self):
        timestamp = self._get_attr('stamp')
        return xep_0082.parse('%sZ' % timestamp) if timestamp else None

    def set_stamp(self, value):
        if isinstance(value, dt.datetime):
            value = value.astimezone(xep_0082.tzutc)
            value = xep_0082.format_datetime(value)
        self._set_attr('stamp', value[0:19].replace('-', ''))

    def get_text(self):
        return self.xml.text

    def set_text(self, value):
        self.xml.text = value

    def del_text(self):
        self.xml.text = ''



# slixmpp.xmlstream.handler.callback
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2011 Nathanael C. Fritz
# :license: MIT, see LICENSE for more details
from __future__ import annotations

from typing import Optional, Callable, Any, TYPE_CHECKING
from slixmpp.xmlstream.handler.base import BaseHandler
from slixmpp.xmlstream.matcher.base import MatcherBase

if TYPE_CHECKING:
    from slixmpp.xmlstream.stanzabase import StanzaBase
    from slixmpp.xmlstream.xmlstream import XMLStream


class Callback(BaseHandler):

    """
    The Callback handler will execute a callback function with
    matched stanzas.

    The handler may execute the callback either during stream
    processing or during the main event loop.

    Callback functions are all executed in the same thread, so be aware if
    you are executing functions that will block for extended periods of
    time. Typically, you should signal your own events using the Slixmpp
    object's :meth:`~slixmpp.xmlstream.xmlstream.XMLStream.event()`
    method to pass the stanza off to a threaded event handler for further
    processing.


    :param string name: The name of the handler.
    :param matcher: A :class:`~slixmpp.xmlstream.matcher.base.MatcherBase`
                    derived object for matching stanza objects.
    :param pointer: The function to execute during callback.
    :param bool once: Indicates if the handler should be used only
                      once. Defaults to False.
    :param bool instream: Indicates if the callback should be executed
                          during stream processing instead of in the
                          main event loop.
    :param stream: The :class:`~slixmpp.xmlstream.xmlstream.XMLStream`
                   instance this handler should monitor.
    """
    _once: bool
    _instream: bool

    def __init__(self, name: str, matcher: MatcherBase,
                 pointer: Callable[[StanzaBase], Any],
                 once: bool = False, instream: bool = False,
                 stream: Optional[XMLStream] = None):
        BaseHandler.__init__(self, name, matcher, stream)
        self._pointer: Callable[[StanzaBase], Any] = pointer
        self._pointer = pointer
        self._once = once
        self._instream = instream

    def prerun(self, payload: StanzaBase) -> None:
        """Execute the callback during stream processing, if
        the callback was created with ``instream=True``.

        :param payload: The matched
            :class:`~slixmpp.xmlstream.stanzabase.StanzaBase` object.
        """
        if self._once:
            self._destroy = True
        if self._instream:
            self.run(payload, True)

    def run(self, payload: StanzaBase, instream: bool = False) -> None:
        """Execute the callback function with the matched stanza payload.

        :param payload: The matched
            :class:`~slixmpp.xmlstream.stanzabase.StanzaBase` object.
        :param bool instream: Force the handler to execute during stream
                              processing. This should only be used by
                              :meth:`prerun()`. Defaults to ``False``.
        """
        if not self._instream or instream:
            self._pointer(payload)
            if self._once:
                self._destroy = True
                del self._pointer




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import asyncio
import logging
import collections

from slixmpp.stanza import Message, Presence, Iq, StreamFeatures
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback, Waiter
from slixmpp.xmlstream.matcher import MatchXPath, MatchMany
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0198 import stanza


log = logging.getLogger(__name__)


MAX_SEQ = 2 ** 32


class XEP_0198(BasePlugin):

    """
    XEP-0198: Stream Management
    """

    name = 'xep_0198'
    description = 'XEP-0198: Stream Management'
    dependencies = set()
    stanza = stanza
    default_config = {
        #: The last ack number received from the server.
        'last_ack': 0,

        #: The number of stanzas to wait between sending ack requests to
        #: the server. Setting this to ``1`` will send an ack request after
        #: every sent stanza. Defaults to ``5``.
        'window': 5,

        #: The stream management ID for the stream. Knowing this value is
        #: required in order to do stream resumption.
        'sm_id': None,

        #: A counter of handled incoming stanzas, mod 2^32.
        'handled': 0,

        #: A counter of unacked outgoing stanzas, mod 2^32.
        'seq': 0,

        #: Control whether or not the ability to resume the stream will be
        #: requested when enabling stream management. Defaults to ``True``.
        'allow_resume': True,

        'order': 10100,
        'resume_order': 9000
    }

    def plugin_init(self):
        """Start the XEP-0198 plugin."""

        # Only enable stream management for non-components,
        # since components do not yet perform feature negotiation.
        if self.xmpp.is_component:
            return

        self.window_counter = self.window

        self.enabled_in = False
        self.enabled_out = False
        self.unacked_queue = collections.deque()

        register_stanza_plugin(StreamFeatures, stanza.StreamManagement)
        self.xmpp.register_stanza(stanza.Enable)
        self.xmpp.register_stanza(stanza.Enabled)
        self.xmpp.register_stanza(stanza.Resume)
        self.xmpp.register_stanza(stanza.Resumed)
        self.xmpp.register_stanza(stanza.Ack)
        self.xmpp.register_stanza(stanza.RequestAck)

        # Register the feature twice because it may be ordered two
        # different ways: enabling after binding and resumption
        # before binding.
        self.xmpp.register_feature('sm',
                self._handle_sm_feature,
                restart=True,
                order=self.order)
        self.xmpp.register_feature('sm',
                self._handle_sm_feature,
                restart=True,
                order=self.resume_order)

        self.xmpp.register_handler(
                Callback('Stream Management Enabled',
                    MatchXPath(stanza.Enabled.tag_name()),
                    self._handle_enabled,
                    instream=True))

        self.xmpp.register_handler(
                Callback('Stream Management Resumed',
                    MatchXPath(stanza.Resumed.tag_name()),
                    self._handle_resumed,
                    instream=True))

        self.xmpp.register_handler(
                Callback('Stream Management Failed',
                    MatchXPath(stanza.Failed.tag_name()),
                    self._handle_failed,
                    instream=True))

        self.xmpp.register_handler(
                Callback('Stream Management Ack',
                    MatchXPath(stanza.Ack.tag_name()),
                    self._handle_ack,
                    instream=True))

        self.xmpp.register_handler(
                Callback('Stream Management Request Ack',
                    MatchXPath(stanza.RequestAck.tag_name()),
                    self._handle_request_ack,
                    instream=True))

        self.xmpp.add_filter('in', self._handle_incoming)
        self.xmpp.add_filter('out_sync', self._handle_outgoing)

        self.xmpp.add_event_handler('disconnected', self.disconnected)
        self.xmpp.add_event_handler('session_end', self.session_end)

    def plugin_end(self):
        if self.xmpp.is_component:
            return

        self.xmpp.unregister_feature('sm', self.order)
        self.xmpp.unregister_feature('sm', self.resume_order)
        self.xmpp.del_event_handler('disconnected', self.disconnected)
        self.xmpp.del_event_handler('session_end', self.session_end)
        self.xmpp.del_filter('in', self._handle_incoming)
        self.xmpp.del_filter('out_sync', self._handle_outgoing)
        self.xmpp.remove_handler('Stream Management Enabled')
        self.xmpp.remove_handler('Stream Management Resumed')
        self.xmpp.remove_handler('Stream Management Failed')
        self.xmpp.remove_handler('Stream Management Ack')
        self.xmpp.remove_handler('Stream Management Request Ack')
        self.xmpp.remove_stanza(stanza.Enable)
        self.xmpp.remove_stanza(stanza.Enabled)
        self.xmpp.remove_stanza(stanza.Resume)
        self.xmpp.remove_stanza(stanza.Resumed)
        self.xmpp.remove_stanza(stanza.Ack)
        self.xmpp.remove_stanza(stanza.RequestAck)

    def disconnected(self, event):
        """Reset enabled state until we can resume/reenable."""
        log.debug("disconnected, disabling SM")
        self.xmpp.event('sm_disabled', event)
        self.enabled_in = False
        self.enabled_out = False

    def session_end(self, event):
        """Reset stream management state."""
        log.debug("session_end, disabling SM")
        self.xmpp.event('sm_disabled', event)
        self.enabled_in = False
        self.enabled_out = False
        self.unacked_queue.clear()
        self.sm_id = None
        self.handled = 0
        self.seq = 0
        self.last_ack = 0

    def send_ack(self):
        """Send the current ack count to the server."""
        if not self.xmpp.transport:
            log.debug('Disconnected: not sending ack')
            return
        ack = stanza.Ack(self.xmpp)
        ack['h'] = self.handled
        self.xmpp.send_raw(str(ack))

    def request_ack(self, e=None):
        """Request an ack from the server."""
        log.debug("requesting ack")
        req = stanza.RequestAck(self.xmpp)
        self.xmpp.send_raw(str(req))

    async def _handle_sm_feature(self, features):
        """
        Enable or resume stream management.

        If no SM-ID is stored, and resource binding has taken place,
        stream management will be enabled.

        If an SM-ID is known, and the server allows resumption, the
        previous stream will be resumed.
        """
        if 'stream_management' in self.xmpp.features:
            # We've already negotiated stream management,
            # so no need to do it again.
            return False
        if self.sm_id and self.allow_resume and 'bind' not in self.xmpp.features:
            resume = stanza.Resume(self.xmpp)
            resume['h'] = self.handled
            resume['previd'] = self.sm_id
            resume.send()
            log.debug("resuming SM")

            # Wait for a response before allowing stream feature processing
            # to continue. The actual result processing will be done in the
            # _handle_resumed() or _handle_failed() methods.
            waiter = Waiter('resumed_or_failed',
                    MatchMany([
                        MatchXPath(stanza.Resumed.tag_name()),
                        MatchXPath(stanza.Failed.tag_name())]))
            self.xmpp.register_handler(waiter)
            result = await waiter.wait()
            if result is not None and result.name == 'resumed':
                return True
            self.xmpp.event("session_end")
        if 'bind' in self.xmpp.features:
            enable = stanza.Enable(self.xmpp)
            enable['resume'] = self.allow_resume
            enable.send()
            log.debug("enabling SM")

            waiter = Waiter('enabled_or_failed',
                    MatchMany([
                        MatchXPath(stanza.Enabled.tag_name()),
                        MatchXPath(stanza.Failed.tag_name())]))
            self.xmpp.register_handler(waiter)
            result = await waiter.wait()
        return False

    def _handle_enabled(self, stanza):
        """Save the SM-ID, if provided.

        Raises an :term:`sm_enabled` event.
        """
        self.xmpp.features.add('stream_management')
        if stanza['id']:
            self.sm_id = stanza['id']
        self.enabled_in = True
        self.handled = 0
        self.xmpp.event('sm_enabled', stanza)
        self.xmpp.end_session_on_disconnect = False

    def _handle_resumed(self, stanza):
        """Finish resuming a stream by resending unacked stanzas.

        Raises a :term:`session_resumed` event.
        """
        self.xmpp.features.add('stream_management')
        self.enabled_in = True
        self._handle_ack(stanza)
        for id, stanza in self.unacked_queue:
            self.xmpp.send(stanza, use_filters=False)
        self.xmpp.event('session_resumed', stanza)
        self.xmpp.end_session_on_disconnect = False

    def _handle_failed(self, stanza):
        """
        Disable and reset any features used since stream management was
        requested (tracked stanzas may have been sent during the interval
        between the enable request and the enabled response).

        Raises an :term:`sm_failed` event.
        """
        self.enabled_in = False
        self.enabled_out = False
        self.unacked_queue.clear()
        self.xmpp.event('sm_failed', stanza)

    def _handle_ack(self, ack):
        """Process a server ack by freeing acked stanzas from the queue.

        Raises a :term:`stanza_acked` event for each acked stanza.
        """
        if ack['h'] == self.last_ack:
            return

        num_acked = (ack['h'] - self.last_ack) % MAX_SEQ
        num_unacked = len(self.unacked_queue)
        log.debug("Ack: %s, Last Ack: %s, " + \
                  "Unacked: %s, Num Acked: %s, " + \
                  "Remaining: %s",
            ack['h'],
            self.last_ack,
            num_unacked,
            num_acked,
            num_unacked - num_acked)
        if num_acked > len(self.unacked_queue) or num_acked < 0:
            log.error('Inconsistent sequence numbers from the server,'
                      ' ignoring and replacing ours with them.')
            num_acked = len(self.unacked_queue)
        for x in range(num_acked):
            seq, stanza = self.unacked_queue.popleft()
            self.xmpp.event('stanza_acked', stanza)
        self.last_ack = ack['h']

    def _handle_request_ack(self, req):
        """Handle an ack request by sending an ack."""
        self.send_ack()

    def _handle_incoming(self, stanza):
        """Increment the handled counter for each inbound stanza."""
        if not self.enabled_in:
            return stanza

        if isinstance(stanza, (Message, Presence, Iq)):
            # Sequence numbers are mod 2^32
            self.handled = (self.handled + 1) % MAX_SEQ
        return stanza

    def _handle_outgoing(self, stanza):
        """Store outgoing stanzas in a queue to be acked."""
        from slixmpp.plugins.xep_0198 import stanza as st
        if isinstance(stanza, (st.Enable, st.Resume)):
            self.enabled_out = True
            # do not clear the queue on resume
            if isinstance(stanza, st.Enable):
                self.unacked_queue.clear()
            log.debug("enabling outgoing SM: %s" % stanza)

        if not self.enabled_out:
            return stanza

        if isinstance(stanza, (Message, Presence, Iq)):
            seq = None
            # Sequence numbers are mod 2^32
            self.seq = (self.seq + 1) % MAX_SEQ
            seq = self.seq
            self.unacked_queue.append((seq, stanza))
            self.window_counter -= 1
            if self.window_counter == 0:
                self.window_counter = self.window
                self.request_ack()
        return stanza



import unittest.mock

from slixmpp.test import SlixTest
# from slixmpp.plugins import xep_0490


class TestMessageDisplaySynchronization(SlixTest):
    def setUp(self):
        self.stream_start(jid="juliet@capulet.lit", plugins={"xep_0490"})

    def test_catch_up(self):
        future = self.xmpp.plugin["xep_0490"].catch_up()
        self.send(  # language=XML
            """
            <iq type="get" to="juliet@capulet.lit" id="1">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <items node="urn:xmpp:mds:displayed:0" />
              </pubsub>
            </iq>
            """
        )
        self.recv(  # language=XML
            """
            <iq type='result'
                to='juliet@capulet.lit/balcony'
                id='1'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <items node='urn:xmpp:mds:displayed:0'>
                  <item id='romeo@montegue.lit'>
                    <displayed xmlns='urn:xmpp:mds:displayed:0'>
                      <stanza-id xmlns='urn:xmpp:sid:0'
                                 id='0f710f2b-52ed-4d52-b928-784dad74a52b'
                                 by='juliet@capulet.lit'/>
                    </displayed>
                  </item>
                  <item id='example@conference.shakespeare.lit'>
                    <displayed xmlns='urn:xmpp:mds:displayed:0'>
                      <stanza-id xmlns='urn:xmpp:sid:0'
                                 id='ca21deaf-812c-48f1-8f16-339a674f2864'
                                 by='example@conference.shakespeare.lit'/>
                    </displayed>
                  </item>
                </items>
              </pubsub>
            </iq>
            """
        )
        iq = future.result()
        item = list(iq["pubsub"]["items"])
        self.assertEqual(item[0]["id"], "romeo@montegue.lit")
        self.assertEqual(
            item[0]["displayed"]["stanza_id"]["id"],
            "0f710f2b-52ed-4d52-b928-784dad74a52b",
        )

        self.assertEqual(item[1]["id"], "example@conference.shakespeare.lit")
        self.assertEqual(
            item[1]["displayed"]["stanza_id"]["id"],
            "ca21deaf-812c-48f1-8f16-339a674f2864",
        )

    def test_flag_chat(self):
        self.xmpp.plugin["xep_0490"].flag_chat(
            "romeo@montegue.lit", "0f710f2b-52ed-4d52-b928-784dad74a52b"
        )
        self.send(  # language=XML
            """
            <iq type='set' id='1'>
              <pubsub xmlns='http://jabber.org/protocol/pubsub'>
                <publish node='urn:xmpp:mds:displayed:0'>
                  <item id='romeo@montegue.lit'>
                    <displayed xmlns='urn:xmpp:mds:displayed:0'>
                      <stanza-id xmlns='urn:xmpp:sid:0' 
                                 id="0f710f2b-52ed-4d52-b928-784dad74a52b" />
                    </displayed>
                  </item>
                </publish>
                <publish-options>
                  <x xmlns='jabber:x:data' type='submit'>
                    <field var='FORM_TYPE' type='hidden'>
                      <value>http://jabber.org/protocol/pubsub#publish-options</value>
                    </field>
                    <field var='pubsub#persist_items'>
                      <value>1</value>
                    </field>
                    <field var='pubsub#max_items'>
                      <value>max</value>
                    </field>
                    <field var='pubsub#send_last_published_item'>
                      <value>never</value>
                    </field>
                    <field var='pubsub#access_model'>
                      <value>whitelist</value>
                    </field>
                  </x>
                </publish-options>
              </pubsub>
            </iq>
            """,
            use_values=False,
        )

    def test_notification(self):
        handler = unittest.mock.Mock()

        self.xmpp.add_event_handler(
            "message_displayed_synchronization_publish", handler
        )
        self.recv(  # language=XML
            """
            <message from='juliet@capulet.lit' to='juliet@capulet.lit/balcony' type='headline' id='new-displayed-pep-event'>
              <event xmlns='http://jabber.org/protocol/pubsub#event'>
                <items node='urn:xmpp:mds:displayed:0'>
                  <item id='romeo@montegue.lit'>
                    <displayed xmlns='urn:xmpp:mds:displayed:0'>
                      <stanza-id xmlns='urn:xmpp:sid:0' by='juliet@capulet.lit' id='0423e3a9-d516-493d-bb06-bee0e51ab9fb'/>
                    </displayed>
                  </item>
                </items>
              </event>
            </message>
            """
        )
        handler.assert_called()
        msg = handler.call_args[0][0]
        self.assertEqual(
            msg["pubsub_event"]["items"]["item"]["id"], "romeo@montegue.lit"
        )
        self.assertEqual(
            msg["pubsub_event"]["items"]["item"]["displayed"]["stanza_id"]["id"],
            "0423e3a9-d516-493d-bb06-bee0e51ab9fb",
        )


suite = unittest.TestLoader().loadTestsFromTestCase(TestMessageDisplaySynchronization)



import logging
import unittest
from slixmpp.test import SlixTest


class TestCaps(SlixTest):
    def setUp(self):
        self.stream_start(plugins=["xep_0115"])

    def testConcurrentSameHash(self):
        """
        Check that we only resolve a given ver string to a disco info once,
        even if we receive several presences with that same ver string
        consecutively.
        """
        self.recv(  # language=XML
            """
            <presence from='romeo@montague.lit/orchard'>
              <c xmlns='http://jabber.org/protocol/caps'
                 hash='sha-1'
                 node='a-node'
                 ver='h0TdMvqNR8FHUfFG1HauOLYZDqE='/>
            </presence>
            """
        )
        self.recv(  # language=XML
            """
            <presence from='i-dont-know-much-shakespeare@montague.lit/orchard'>
              <c xmlns='http://jabber.org/protocol/caps'
                 hash='sha-1'
                 node='a-node'
                 ver='h0TdMvqNR8FHUfFG1HauOLYZDqE='/>
            </presence>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:client"
                id="1"
                to="romeo@montague.lit/orchard"
                type="get">
              <query xmlns="http://jabber.org/protocol/disco#info"
                     node="a-node#h0TdMvqNR8FHUfFG1HauOLYZDqE="/>
            </iq>
            """
        )
        self.send(None)
        self.recv(  # language=XML
            """
            <iq from='romeo@montague.lit/orchard'
                id='1'
                type='result'>
              <query xmlns='http://jabber.org/protocol/disco#info'
                     node='a-nodes#h0TdMvqNR8FHUfFG1HauOLYZDqE='>
                <identity category='client' name='a client' type='pc'/>
                <feature var='http://jabber.org/protocol/caps'/>
              </query>
            </iq>
            """
        )
        self.send(None)
        self.assertTrue(
            self.xmpp["xep_0030"].supports(
                "romeo@montague.lit/orchard", "http://jabber.org/protocol/caps"
            )
        )
        self.assertTrue(
            self.xmpp["xep_0030"].supports(
                "i-dont-know-much-shakespeare@montague.lit/orchard",
                "http://jabber.org/protocol/caps",
            )
        )


logging.basicConfig(level=logging.DEBUG)
suite = unittest.TestLoader().loadTestsFromTestCase(TestCaps)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8 et ts=4 sts=4 sw=4
#
# Copyright © 2022 Maxime “pep” Buquet <pep@bouah.net>
#
# Distributed under terms of the GPLv3+ license.

"""
    Tests for XEP-0454 (OMEMO Media Sharing) plugin.
"""

import unittest
from io import BytesIO
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0454 import XEP_0454


class TestMediaSharing(SlixTest):

    def testEncryptDecryptSmall(self):
        plain = b'qwertyuiop'
        ciphertext, fragment = XEP_0454.encrypt(input_file=BytesIO(plain))
        result = XEP_0454.decrypt(BytesIO(ciphertext), fragment)

        self.assertEqual(plain, result)

    def testEncryptDecrypt(self):
        plain = b'a' * 4096 + b'qwertyuiop'
        ciphertext, fragment = XEP_0454.encrypt(input_file=BytesIO(plain))
        result = XEP_0454.decrypt(BytesIO(ciphertext), fragment)

        self.assertEqual(plain, result)

    def testFormatURL(self):
        url = 'https://foo.bar'
        fragment = 'a' * 88
        result = XEP_0454.format_url(url, fragment)
        self.assertEqual('aesgcm://foo.bar#' + 'a' * 88, result)

suite = unittest.TestLoader().loadTestsFromTestCase(TestMediaSharing)



#!/usr/bin/env python3

import sys
import logging
from getpass import getpass
from argparse import ArgumentParser

try:
    from appscript import *
except ImportError:
    print('This demo requires the appscript package to interact with iTunes.')
    sys.exit()

import asyncio
from slixmpp import ClientXMPP


class TuneBot(ClientXMPP):

    def __init__(self, jid, password):
        super().__init__(jid, password)

        # Check for the current song every 5 seconds.
        self.schedule('Check Current Tune', 5, self._update_tune, repeat=True)

        self.add_event_handler('session_start', self.start)
        self.add_event_handler('user_tune_publish', self.user_tune_publish)

        self.register_plugin('xep_0004')
        self.register_plugin('xep_0030')
        self.register_plugin('xep_0060')
        self.register_plugin('xep_0115')
        self.register_plugin('xep_0118')
        self.register_plugin('xep_0128')
        self.register_plugin('xep_0163')

        self.current_tune = None

    async def start(self, event):
        self.send_presence()
        await self.get_roster()
        self['xep_0115'].update_caps()

    def _update_tune(self):
        itunes_count = app('System Events').processes[its.name == 'iTunes'].count()
        if itunes_count > 0:
            iTunes = app('iTunes')
            if iTunes.player_state.get() == k.playing:
                track = iTunes.current_track.get()
                length = track.time.get()
                if ':' in length:
                    minutes, secs = map(int, length.split(':'))
                    secs += minutes * 60
                else:
                    secs = int(length)

                artist = track.artist.get()
                title = track.name.get()
                source = track.album.get()
                rating = track.rating.get() / 10

                tune = (artist, secs, rating, source, title)
                if tune != self.current_tune:
                    self.current_tune = tune

                    # We have a new song playing, so publish it.
                    self['xep_0118'].publish_tune(
                            artist=artist,
                            length=secs,
                            title=title,
                            rating=rating,
                            source=source)
            else:
                # No song is playing, clear the user tune.
                tune = None
                if tune != self.current_tune:
                    self.current_tune = tune
                    self['xep_0118'].stop()

    def user_tune_publish(self, msg):
        tune = msg['pubsub_event']['items']['item']['tune']
        print("%s is listening to: %s" % (msg['from'], tune['title']))


if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()

    # Output verbosity options.
    parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                        action="store_const", dest="loglevel",
                        const=logging.ERROR, default=logging.INFO)
    parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                        action="store_const", dest="loglevel",
                        const=logging.DEBUG, default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")

    xmpp = TuneBot(args.jid, args.password)

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_forever()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2015 Emmanuel Gil Peyrot
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0070.stanza import Confirm
from slixmpp.plugins.xep_0070.confirm import XEP_0070


register_plugin(XEP_0070)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.google.nosave import stanza
from slixmpp.plugins.google.nosave.nosave import GoogleNoSave



import asyncio
import unittest
from uuid import uuid4
from slixmpp import JID
from slixmpp.test.integration import SlixIntegration
from slixmpp.plugins.xep_0410 import PingStatus

UNIQUE = uuid4().hex


class TestSelfPing(SlixIntegration):

    async def asyncSetUp(self):
        self.mucserver = self.envjid('CI_MUC_SERVER', default='chat.jabberfr.org')
        self.muc = JID('%s@%s' % (UNIQUE, self.mucserver))
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        self.register_plugins(['xep_0410'], [{'ping_interval': 2}])
        await self.connect_clients()

    async def config_room(self):
        self.clients[0]['xep_0045'].join_muc(self.muc, 'client1')
        presence = await self.clients[0].wait_until('muc::%s::got_online' % self.muc)
        config = await self.clients[0]['xep_0045'].get_room_config(self.muc)
        values = config.get_values()
        values['muc#roomconfig_persistentroom'] = False
        values['muc#roomconfig_membersonly'] = True
        config['values'] = values
        config.reply()
        config = await self.clients[0]['xep_0045'].set_room_config(self.muc, config)
        await self.clients[0]['xep_0045'].send_affiliation_list(
            self.muc,
            [
                (self.clients[1].boundjid.bare, 'member'),
            ],
        )

    async def test_presence_monitor(self):
        """Check that the ping status gets updated on room changes"""
        await self.config_room()
        self.clients[1]['xep_0045'].join_muc(self.muc, 'client2')
        full = JID(self.muc)
        full.resource = 'client2'
        self.clients[1]['xep_0410'].enable_self_ping(full)
        await self.clients[1].wait_until('muc::%s::got_online' % self.muc)
        await asyncio.sleep(3)
        self.assertEqual(
            self.clients[1]['xep_0410'].get_ping_status(full),
            PingStatus.JOINED,
        )
        t = asyncio.create_task
        _, pending = await asyncio.wait(
            [
                t(self.clients[0]['xep_0045'].set_role(self.muc, 'client2', 'none')),
                t(self.clients[0].wait_until('muc::%s::got_offline' % self.muc)),
                t(self.clients[1].wait_until('muc_ping_changed')),
            ],
            timeout=10,
        )
        self.assertEqual(pending, set())
        self.assertEqual(
            self.clients[1]['xep_0410'].get_ping_status(full),
            PingStatus.DISCONNECTED,
        )
        self.clients[1]['xep_0045'].join_muc(self.muc, 'client2')
        await asyncio.wait(
            [
                t(self.clients[1].wait_until('muc::%s::got_online' % self.muc)),
                t(self.clients[1].wait_until('muc_ping_changed')),
                t(asyncio.sleep(3))],
            timeout=10,
        )
        self.assertEqual(
            self.clients[1]['xep_0410'].get_ping_status(full),
            PingStatus.JOINED,
        )


suite = unittest.TestLoader().loadTestsFromTestCase(TestSelfPing)



# -*- encoding:utf-8 -*-
from __future__ import unicode_literals

import unittest
from slixmpp.exceptions import IqTimeout
from slixmpp.test import SlixTest
import time
import threading


class TestStreamRoster(SlixTest):
    """
    Test handling roster updates.
    """

    def testGetRoster(self):
        """Test handling roster requests."""
        self.stream_start(mode='client', jid='tester@localhost')

        roster_updates = []

        self.xmpp.add_event_handler('roster_update', roster_updates.append)

        self.xmpp.get_roster()

        self.send("""
          <iq type="get" id="1">
            <query xmlns="jabber:iq:roster" />
          </iq>
        """)
        self.recv("""
          <iq to='tester@localhost' type="result" id="1">
            <query xmlns="jabber:iq:roster">
              <item jid="user@localhost"
                    name="User"
                    subscription="from"
                    ask="subscribe">
                <group>Friends</group>
                <group>Examples</group>
              </item>
            </query>
          </iq>
        """)

        self.check_roster('tester@localhost', 'user@localhost',
                          name='User',
                          subscription='from',
                          afrom=True,
                          pending_out=True,
                          groups=['Friends', 'Examples'])

        self.assertTrue(len(roster_updates) == 1,
                "Wrong number of roster_update events fired: %s (should be 1)" % len(roster_updates))

    def testRosterSet(self):
        """Test handling pushed roster updates."""
        self.stream_start(mode='client')
        events = []

        def roster_update(e):
            events.append('roster_update')

        self.xmpp.add_event_handler('roster_update', roster_update)

        self.recv("""
          <iq to='tester@localhost' type="set" id="1">
            <query xmlns="jabber:iq:roster">
              <item jid="user@localhost"
                    name="User"
                    subscription="both">
                <group>Friends</group>
                <group>Examples</group>
              </item>
            </query>
          </iq>
        """)
        self.send("""
          <iq type="result" id="1">
            <query xmlns="jabber:iq:roster" />
          </iq>
        """)

        self.check_roster('tester@localhost', 'user@localhost',
                          name='User',
                          subscription='both',
                          groups=['Friends', 'Examples'])


        self.assertTrue('roster_update' in events,
                "Roster updated event not triggered: %s" % events)

    def testRosterPushRemove(self):
        """Test handling roster item removal updates."""
        self.stream_start(mode='client')
        events = []

        # Add roster item
        self.recv("""
          <iq to='tester@localhost' type="set" id="1">
            <query xmlns="jabber:iq:roster">
              <item jid="user@localhost"
                    name="User"
                    subscription="both">
                <group>Friends</group>
                <group>Examples</group>
              </item>
            </query>
          </iq>
        """)
        self.send("""
          <iq type="result" id="1">
            <query xmlns="jabber:iq:roster" />
          </iq>
        """)

        self.assertTrue('user@localhost' in self.xmpp.client_roster)

        # Receive item remove push
        self.recv("""
          <iq to='tester@localhost' type="set" id="1">
            <query xmlns="jabber:iq:roster">
              <item jid="user@localhost"
                    subscription="remove">
              </item>
            </query>
          </iq>
        """)
        self.send("""
          <iq type="result" id="1">
            <query xmlns="jabber:iq:roster" />
          </iq>
        """)

        self.assertTrue('user@localhost' not in self.xmpp.client_roster)

    def testUnauthorizedRosterPush(self):
        """Test rejecting a roster push from an unauthorized source."""
        self.stream_start()
        self.recv("""
          <iq to='tester@localhost' from="malicious_user@localhost"
              type="set" id="1">
            <query xmlns="jabber:iq:roster">
              <item jid="user@localhost"
                    name="User"
                    subscription="both">
                <group>Friends</group>
                <group>Examples</group>
              </item>
            </query>
          </iq>
        """)
        self.send("""
          <iq to="malicious_user@localhost" type="error" id="1">
            <error type="cancel" code="503">
              <service-unavailable xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
            </error>
          </iq>
        """)

    def testRosterCallback(self):
        """Test handling a roster request callback."""
        self.stream_start()
        events = []

        def roster_callback(iq):
            events.append('roster_callback')

        self.xmpp.get_roster(callback=roster_callback)

        self.send("""
          <iq type="get" id="1">
            <query xmlns="jabber:iq:roster" />
          </iq>
        """)
        self.recv("""
          <iq type="result" id="1">
            <query xmlns="jabber:iq:roster">
              <item jid="user@localhost"
                    name="User"
                    subscription="both">
                <group>Friends</group>
                <group>Examples</group>
              </item>
            </query>
          </iq>
        """)

        self.assertTrue(events == ['roster_callback'],
                 "Roster timeout event not triggered: %s." % events)

    def testRosterUnicode(self):
        """Test that JIDs with Unicode values are handled properly."""
        self.stream_start(plugins=[])
        self.recv("""
          <iq to="tester@localhost" type="set" id="1">
            <query xmlns="jabber:iq:roster">
              <item jid="andré@foo" subscription="both">
                <group>Unicode</group>
              </item>
            </query>
          </iq>
        """)

        self.check_roster('tester@localhost', 'andré@foo',
                          subscription='both',
                          groups=['Unicode'])

        jids = list(self.xmpp.client_roster.keys())
        self.assertTrue(jids == ['andré@foo'],
                 "Too many roster entries found: %s" % jids)

        self.recv("""
          <presence to="tester@localhost" from="andré@foo/bar">
            <show>away</show>
            <status>Testing</status>
          </presence>
        """)

        result = self.xmpp.client_roster['andré@foo'].resources
        expected = {'bar': {'status':'Testing',
                            'show':'away',
                            'priority':0}}
        self.assertTrue(result == expected,
                "Unexpected roster values: %s" % result)

    def testSendLastPresence(self):
        """Test that sending the last presence works."""
        self.stream_start(plugins=[])
        self.xmpp.send_presence(pshow='dnd')
        self.xmpp.auto_authorize = True
        self.xmpp.auto_subscribe = True

        self.send("""
          <presence>
            <show>dnd</show>
          </presence>
        """)

        self.recv("""
          <presence from="user@localhost"
                    to="tester@localhost"
                    type="subscribe" />
        """)

        self.send("""
          <presence to="user@localhost"
                    type="subscribed" />
        """)

        self.send("""
          <presence to="user@localhost">
            <show>dnd</show>
          </presence>
        """)

    def testUnsupportedRosterVer(self):
        """Test working with a server without roster versioning."""
        self.stream_start()
        self.assertTrue('rosterver' not in self.xmpp.features)

        self.xmpp.get_roster()

        self.send("""
          <iq type="get" id="1">
            <query xmlns="jabber:iq:roster" />
          </iq>
        """)
        self.recv("""
          <iq to="tester@localhost" type="result" id="1" />
        """)

    def testBootstrapRosterVer(self):
        """Test bootstrapping with roster versioning."""
        self.stream_start()
        self.xmpp.features.add('rosterver')
        self.xmpp.client_roster.version = ''

        self.xmpp.get_roster()

        self.send("""
          <iq type="get" id="1">
            <query xmlns="jabber:iq:roster" ver="" />
          </iq>
        """)
        self.recv("""
          <iq to="tester@localhost" type="result" id="1" />
        """)


    def testExistingRosterVer(self):
        """Test using a stored roster version."""
        self.stream_start()
        self.xmpp.features.add('rosterver')
        self.xmpp.client_roster.version = '42'

        self.xmpp.get_roster()

        self.send("""
          <iq type="get" id="1">
            <query xmlns="jabber:iq:roster" ver="42" />
          </iq>
        """)
        self.recv("""
          <iq to="tester@localhost" type="result" id="1" />
        """)


suite = unittest.TestLoader().loadTestsFromTestCase(TestStreamRoster)



import unittest
from slixmpp.test.integration import SlixIntegration


class TestRetract(SlixIntegration):
    async def asyncSetUp(self):
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        self.register_plugins(['xep_0424'])
        await self.connect_clients()

    async def test_retract_msg(self):
        """Try to retract a message"""
        self.clients[0]['xep_0424'].send_retraction(
            self.clients[1].boundjid.full,
            id='toto',
            fallback_text='Twas a mistake',
        )
        msg = await self.clients[1].wait_until('message_retract')
        self.assertEqual(msg['retract']['id'], 'toto')

suite = unittest.TestLoader().loadTestsFromTestCase(TestRetract)



import unittest
from slixmpp import Message
from slixmpp.test import SlixTest
from slixmpp.xmlstream import ET
from slixmpp.plugins.xep_0422 import stanza


class TestFastening(SlixTest):

    def setUp(self):
        stanza.register_plugins()

    def testFastenExternal(self):
        message = Message()
        message['apply_to']['id'] = 'some-id'
        message['apply_to'].xml.append(
            ET.fromstring('<test xmlns="urn:tmp:test">Test</test>')
        )
        message['apply_to']['external']['name'] = 'body'
        message['body'] = 'Toto'

        self.check(message, """
<message>
  <apply-to xmlns="urn:xmpp:fasten:0" id="some-id">
      <test xmlns="urn:tmp:test">Test</test>
      <external name='body'/>
  </apply-to>
  <body>Toto</body>
</message>
        """, use_values=False)


suite = unittest.TestLoader().loadTestsFromTestCase(TestFastening)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ET, ElementBase, JID


class BlockList(ElementBase):
    name = 'blocklist'
    namespace = 'urn:xmpp:blocking'
    plugin_attrib = 'blocklist'
    interfaces = {'items'}

    def get_items(self):
        result = set()
        items = self.xml.findall('{%s}item' % self.namespace)
        if items is not None:
            for item in items:
                jid = JID(item.attrib.get('jid', ''))
                if jid:
                    result.add(jid)
        return result

    def set_items(self, values):
        self.del_items()
        for jid in values:
            if jid:
                item = ET.Element('{%s}item' % self.namespace)
                item.attrib['jid'] = JID(jid).full
                self.xml.append(item)

    def del_items(self):
        items = self.xml.findall('{%s}item' % self.namespace)
        if items is not None:
            for item in items:
                self.xml.remove(item)


class Block(BlockList):
    name = 'block'
    plugin_attrib = 'block'


class Unblock(BlockList):
    name = 'unblock'
    plugin_attrib = 'unblock'



from slixmpp import Presence
from slixmpp.xmlstream import ElementBase, register_stanza_plugin
from typing import List, Tuple

NS = 'urn:xmpp:hats:0'


class Hats(ElementBase):
    """
    Hats element, container for multiple hats:

    .. code-block::xml


      <hats xmlns='urn:xmpp:hats:0'>
        <hat title='Host' uri='http://schemas.example.com/hats#host' xml:lang='en-us'>
            <badge xmlns="urn:example:badges" fgcolor="#000000" bgcolor="#58C5BA"/>
        </hat>
        <hat title='Presenter' uri='http://schemas.example.com/hats#presenter' xml:lang='en-us'>
            <badge xmlns="urn:example:badges" fgcolor="#000000" bgcolor="#EC0524"/>
        </hat>
      </hats>

    """

    name = 'hats'
    namespace = NS
    plugin_attrib = 'hats'

    def add_hats(self, data: List[Tuple[str, str]]) -> None:
        for uri, title in data:
            hat = Hat()
            hat["uri"] = uri
            hat["title"] = title
            self.append(hat)


class Hat(ElementBase):
    """
    Hat element, has a title and url, may contain arbitrary sub-elements.

    .. code-block::xml

        <hat title='Host' uri='http://schemas.example.com/hats#host' xml:lang='en-us'>
            <badge xmlns="urn:example:badges" fgcolor="#000000" bgcolor="#58C5BA"/>
        </hat>

    """
    name = 'hat'
    plugin_attrib = 'hat'
    namespace = NS
    interfaces = {'title', 'uri'}
    plugin_multi_attrib = "hats"


def register_plugin() -> None:
    register_stanza_plugin(Hats, Hat, iterable=True)
    register_stanza_plugin(Presence, Hats)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase, ET, register_stanza_plugin


class Certs(ElementBase):
    name = 'items'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'sasl_certs'
    interfaces = set()


class CertItem(ElementBase):
    name = 'item'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'item'
    plugin_multi_attrib = 'items'
    interfaces = {'name', 'x509cert', 'users'}
    sub_interfaces = {'name', 'x509cert'}

    def get_users(self):
        resources = self.xml.findall('{%s}users/{%s}resource' % (
            self.namespace, self.namespace))
        return {res.text for res in resources}

    def set_users(self, values):
        users = self.xml.find('{%s}users' % self.namespace)
        if users is None:
            users = ET.Element('{%s}users' % self.namespace)
            self.xml.append(users)
        for resource in values:
            res = ET.Element('{%s}resource' % self.namespace)
            res.text = resource
            users.append(res)

    def del_users(self):
        users = self.xml.find('{%s}users' % self.namespace)
        if users is not None:
            self.xml.remove(users)


class AppendCert(ElementBase):
    name = 'append'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'sasl_cert_append'
    interfaces = {'name', 'x509cert', 'cert_management'}
    sub_interfaces = {'name', 'x509cert'}

    def get_cert_management(self):
        manage = self.xml.find('{%s}no-cert-management' % self.namespace)
        return manage is None

    def set_cert_management(self, value):
        self.del_cert_management()
        if not value:
            manage = ET.Element('{%s}no-cert-management' % self.namespace)
            self.xml.append(manage)

    def del_cert_management(self):
        manage = self.xml.find('{%s}no-cert-management' % self.namespace)
        if manage is not None:
            self.xml.remove(manage)


class DisableCert(ElementBase):
    name = 'disable'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'sasl_cert_disable'
    interfaces = {'name'}
    sub_interfaces = interfaces


class RevokeCert(ElementBase):
    name = 'revoke'
    namespace = 'urn:xmpp:saslcert:1'
    plugin_attrib = 'sasl_cert_revoke'
    interfaces = {'name'}
    sub_interfaces = interfaces


register_stanza_plugin(Certs, CertItem, iterable=True)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0196 import stanza
from slixmpp.plugins.xep_0196.stanza import UserGaming
from slixmpp.plugins.xep_0196.user_gaming import XEP_0196


register_plugin(XEP_0196)



import datetime
import unittest

from slixmpp import Iq
from slixmpp.test import SlixTest

from slixmpp.plugins.xep_0292 import stanza


REF = """
<iq>
    <vcard xmlns='urn:ietf:params:xml:ns:vcard-4.0'>
        <fn>
            <text>Full Name</text>
        </fn>
        <n><given>Full</given><surname>Name</surname></n>
        <nickname>
            <text>some nick</text>
        </nickname>
        <bday>
            <date>1984-05-21</date>
        </bday>
        <url>
            <uri>https://nicoco.fr</uri>
        </url>
        <note>
            <text>About me</text>
        </note>
        <impp>
            <uri>xmpp:test@localhost</uri>
        </impp>
        <email>
            <text>test@gmail.com</text>
        </email>
        <tel>
            <parameters>
                <type><text>work</text></type>
            </parameters>
            <uri>tel:+555</uri>
        </tel>
        <adr>
            <locality>Nice</locality>
            <country>France</country>
        </adr>
    </vcard>
</iq>
"""


class TestVcard(SlixTest):
    def test_basic_interfaces(self):
        iq = Iq()
        x = iq["vcard"]

        x["fn"]["text"] = "Full Name"
        x["nickname"]["text"] = "some nick"
        x["n"]["given"] = "Full"
        x["n"]["surname"] = "Name"
        x["bday"]["date"] = datetime.date(1984, 5, 21)
        x["note"]["text"] = "About me"
        x["url"]["uri"] = "https://nicoco.fr"
        x["impp"]["uri"] = "xmpp:test@localhost"
        x["email"]["text"] = "test@gmail.com"

        x["tel"]["uri"] = "tel:+555"
        x["tel"]["parameters"]["type_"]["text"] = "work"
        x["adr"]["locality"] = "Nice"
        x["adr"]["country"] = "France"

        self.check(iq, REF, use_values=False)

    def test_easy_interface(self):
        iq = Iq()
        x: stanza.VCard4 = iq["vcard"]

        x["full_name"] = "Full Name"
        x["given"] = "Full"
        x["surname"] = "Name"
        x["birthday"] = datetime.date(1984, 5, 21)
        x.add_nickname("some nick")
        x.add_note("About me")
        x.add_url("https://nicoco.fr")
        x.add_impp("xmpp:test@localhost")
        x.add_email("test@gmail.com")
        x.add_tel("+555", "work")
        x.add_address("France", "Nice")

        self.check(iq, REF, use_values=False)

    def test_2_phones(self):
        vcard = stanza.VCard4()
        tel1 = stanza.Tel()
        tel1["parameters"]["type_"]["text"] = "work"
        tel1["uri"] = "tel:+555"
        tel2 = stanza.Tel()
        tel2["parameters"]["type_"]["text"] = "devil"
        tel2["uri"] = "tel:+666"
        vcard.append(tel1)
        vcard.append(tel2)
        self.check(
            vcard,
            """
            <vcard xmlns='urn:ietf:params:xml:ns:vcard-4.0'>
                <tel>
                    <parameters>
                        <type><text>work</text></type>
                    </parameters>
                    <uri>tel:+555</uri>
                </tel>
                <tel>
                    <parameters>
                        <type><text>devil</text></type>
                    </parameters>
                    <uri>tel:+666</uri>
                </tel>
            </vcard>
            """,
            use_values=False
        )

suite = unittest.TestLoader().loadTestsFromTestCase(TestVcard)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase, ET


class Attention(ElementBase):

    """
    """

    name = 'attention'
    namespace = 'urn:xmpp:attention:0'
    plugin_attrib = 'attention'
    interfaces = {'attention'}
    is_extension = True

    def setup(self, xml):
        return True

    def set_attention(self, value):
        if value:
            xml = ET.Element(self.tag_name())
            self.parent().xml.append(xml)
        else:
            self.del_attention()

    def get_attention(self):
        xml = self.parent().xml.find(self.tag_name())
        return xml is not None

    def del_attention(self):
        xml = self.parent().xml.find(self.tag_name())
        if xml is not None:
            self.parent().xml.remove(xml)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import StanzaBase, ElementBase
from typing import ClassVar, Dict, Type


class StreamFeatures(StanzaBase):

    """
    """

    name = 'features'
    namespace = 'http://etherx.jabber.org/streams'
    interfaces = {'features', 'required', 'optional'}
    sub_interfaces = interfaces
    plugin_attrib_map: ClassVar[Dict[str, Type[ElementBase]]] = {}
    plugin_tag_map: ClassVar[Dict[str, Type[ElementBase]]] = {}

    def setup(self, xml):
        StanzaBase.setup(self, xml)
        self.values = self.values

    def get_features(self):
        """
        """
        features = {}
        for (name, lang), plugin in self.plugins.items():
            features[name] = plugin
        return features

    def set_features(self, value):
        """
        """
        pass

    def del_features(self):
        """
        """
        pass

    def get_required(self):
        """
        """
        features = self['features']
        return [f for n, f in features.items() if f['required']]

    def get_optional(self):
        """
        """
        features = self['features']
        return [f for n, f in features.items() if not f['required']]




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Iq, Message
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.google.nosave import stanza


class GoogleNoSave(BasePlugin):

    """
    Google: Off the Record Chats

    NOTE: This is NOT an encryption method.

    Also see <https://developers.google.com/talk/jep_extensions/otr>.
    """

    name = 'google_nosave'
    description = 'Google: Off the Record Chats'
    dependencies = set(['google_settings'])
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, stanza.NoSave)
        register_stanza_plugin(Iq, stanza.NoSaveQuery)

        self.xmpp.register_handler(
                Callback('Google Nosave',
                    StanzaPath('iq@type=set/google_nosave'),
                    self._handle_nosave_change))

    def plugin_end(self):
        self.xmpp.remove_handler('Google Nosave')

    def enable(self, jid=None, timeout=None, callback=None):
        if jid is None:
            self.xmpp['google_settings'].update({'archiving_enabled': False},
                    timeout=timeout, callback=callback)
        else:
            iq = self.xmpp.Iq()
            iq['type'] = 'set'
            iq['google_nosave']['item']['jid'] = jid
            iq['google_nosave']['item']['value'] = True
            return iq.send(timeout=timeout, callback=callback)

    def disable(self, jid=None, timeout=None, callback=None):
        if jid is None:
            self.xmpp['google_settings'].update({'archiving_enabled': True},
                    timeout=timeout, callback=callback)
        else:
            iq = self.xmpp.Iq()
            iq['type'] = 'set'
            iq['google_nosave']['item']['jid'] = jid
            iq['google_nosave']['item']['value'] = False
            return iq.send(timeout=timeout, callback=callback)

    def get(self, timeout=None, callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq.enable('google_nosave')
        return iq.send(timeout=timeout, callback=callback)

    def _handle_nosave_change(self, iq):
        reply = self.xmpp.Iq()
        reply['type'] = 'result'
        reply['id'] = iq['id']
        reply['to'] = iq['from']
        reply.send()
        self.xmpp.event('google_nosave_change', iq)



import unittest
from slixmpp.test import SlixTest
from slixmpp.xmlstream.stanzabase import ET


class TestIqStanzas(SlixTest):

    def testSetup(self):
        """Test initializing default Iq values."""
        iq = self.Iq()
        self.check(iq, """
          <iq id="0" />
        """)

    def testPayload(self):
        """Test setting Iq stanza payload."""
        iq = self.Iq()
        iq.set_payload(ET.Element('{test}tester'))
        self.check(iq, """
          <iq id="0">
            <tester xmlns="test" />
          </iq>
        """, use_values=False)


    def testUnhandled(self):
        """Test behavior for Iq.unhandled."""
        self.stream_start()
        self.recv("""
          <iq id="test" type="get">
            <query xmlns="test" />
           </iq>
        """)

        iq = self.Iq()
        iq['id'] = 'test'
        iq['error']['condition'] = 'feature-not-implemented'
        iq['error']['text'] = 'No handlers registered for this request.'

        self.send(iq, """
          <iq id="test" type="error">
            <error type="cancel">
              <feature-not-implemented xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
              <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">
                No handlers registered for this request.
              </text>
            </error>
          </iq>
        """)

    def testQuery(self):
        """Test modifying query element of Iq stanzas."""
        iq = self.Iq()

        iq['query'] = 'query_ns'
        self.check(iq, """
          <iq id="0">
            <query xmlns="query_ns" />
          </iq>
        """)

        iq['query'] = 'query_ns2'
        self.check(iq, """
          <iq id="0">
            <query xmlns="query_ns2" />
          </iq>
        """)

        self.assertTrue(iq['query'] == 'query_ns2', "Query namespace doesn't match")

        del iq['query']
        self.check(iq, """
          <iq id="0" />
        """)

    def testReply(self):
        """Test setting proper result type in Iq replies."""
        iq = self.Iq()
        iq['to'] = 'user@localhost'
        iq['type'] = 'get'
        iq = iq.reply()

        self.check(iq, """
          <iq id="0" type="result" />
        """)

suite = unittest.TestLoader().loadTestsFromTestCase(TestIqStanzas)



# slixmpp: The Slick XMPP Library
# Copyright (C) 2016 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
import logging
from typing import Optional

from slixmpp import Message, JID
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.xep_0333 import stanza, Markable, Received, Displayed, Acknowledged

log = logging.getLogger(__name__)

class XEP_0333(BasePlugin):

    name = 'xep_0333'
    description = 'XEP-0333: Chat Markers'
    stanza = stanza
    dependencies = {'xep_0030'}

    def plugin_init(self):
        register_stanza_plugin(Message, Markable)
        register_stanza_plugin(Message, Received)
        register_stanza_plugin(Message, Displayed)
        register_stanza_plugin(Message, Acknowledged)

        self.xmpp.register_handler(
            Callback('Received Chat Marker',
                StanzaPath('message/received'),
                self._handle_received))
        self.xmpp.register_handler(
            Callback('Displayed Chat Marker',
                StanzaPath('message/displayed'),
                self._handle_displayed))
        self.xmpp.register_handler(
            Callback('Acknowledged Chat Marker',
                StanzaPath('message/acknowledged'),
                self._handle_acknowledged))

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(stanza.NS)

    def plugin_end(self):
        self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.NS)

    def _handle_received(self, message):
        self.xmpp.event('marker_received', message)
        self.xmpp.event('marker', message)

    def _handle_displayed(self, message):
        self.xmpp.event('marker_displayed', message)
        self.xmpp.event('marker', message)

    def _handle_acknowledged(self, message):
        self.xmpp.event('marker_acknowledged', message)
        self.xmpp.event('marker', message)

    def send_marker(self, mto: JID, id: str, marker: str,
                    thread: Optional[str] = None, *,
                    mfrom: Optional[JID] = None):
        """
        Send a chat marker.

        :param JID mto: recipient of the marker
        :param str id: Identifier of the marked message
        :param str marker: Marker to send (one of
            displayed, received, or acknowledged)
        :param str thread: Message thread
        :param str mfrom: Use a specific JID to send the message
        """
        if marker not in ('displayed', 'received', 'acknowledged'):
            raise ValueError('Invalid marker: %s' % marker)
        msg = self.xmpp.make_message(mto=mto, mfrom=mfrom)
        if thread:
            msg['thread'] = thread
        msg[marker]['id'] = id
        msg.send()



import unittest
from slixmpp.test.integration import SlixIntegration


class TestPing(SlixIntegration):
    async def asyncSetUp(self):
        await super().asyncSetUp()
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.register_plugins(['xep_0199'])
        await self.connect_clients()

    async def test_ping(self):
        """Check we can ping our own server"""
        rtt = await self.clients[0]['xep_0199'].ping()
        self.assertGreater(10, rtt)


suite = unittest.TestLoader().loadTestsFromTestCase(TestPing)



import asyncio
import unittest
from slixmpp.test.integration import SlixIntegration
from slixmpp import Message


class TestSlowFilter(SlixIntegration):
    async def asyncSetUp(self):
        await super().asyncSetUp()
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        await self.connect_clients()

    async def test_filters(self):
        """Make sure filters work"""
        def add_a(stanza):
            if isinstance(stanza, Message):
                stanza['body'] = stanza['body'] + ' a'
            return stanza

        async def add_b(stanza):
            if isinstance(stanza, Message):
                stanza['body'] = stanza['body'] + ' b'
            return stanza

        async def add_c_wait(stanza):
            if isinstance(stanza, Message):
                await asyncio.sleep(2)
                stanza['body'] = stanza['body'] + ' c'
            return stanza
        self.clients[0].add_filter('out', add_a)
        self.clients[0].add_filter('out', add_b)
        self.clients[0].add_filter('out', add_c_wait)
        body = 'Msg body'
        msg = self.clients[0].make_message(
            mto=self.clients[1].boundjid, mbody=body,
        )
        msg.send()
        message = await self.clients[1].wait_until('message')
        self.assertEqual(message['body'], body + ' a b c')


suite = unittest.TestLoader().loadTestsFromTestCase(TestSlowFilter)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from asyncio import Future

from slixmpp import JID
from typing import Dict, List, Optional, Callable
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0152 import stanza, Reachability
from slixmpp.plugins.xep_0004 import Form


log = logging.getLogger(__name__)


class XEP_0152(BasePlugin):

    """
    XEP-0152: Reachability Addresses
    """

    name = 'xep_0152'
    description = 'XEP-0152: Reachability Addresses'
    dependencies = {'xep_0163'}
    stanza = stanza

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=Reachability.namespace)
        self.xmpp['xep_0163'].remove_interest(Reachability.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0163'].register_pep('reachability', Reachability)

    def publish_reachability(self, addresses: List[Dict[str, str]],
                             **pubsubkwargs) -> Future:
        """
        Publish alternative addresses where the user can be reached.

        :param addresses: A list of dictionaries containing the URI and
                          optional description for each address.
        """
        if not isinstance(addresses, (list, tuple)):
            addresses = [addresses]
        reach = Reachability()
        for address in addresses:
            if not hasattr(address, 'items'):
                address = {'uri': address}

            addr = stanza.Address()
            for key, val in address.items():
                addr[key] = val
            reach.append(addr)
        return self.xmpp['xep_0163'].publish(
            reach,
            node=Reachability.namespace,
            **pubsubkwargs
        )

    def stop(self, **pubsubkwargs) -> Future:
        """
        Clear existing user activity information to stop notifications.
        """
        reach = Reachability()
        return self.xmpp['xep_0163'].publish(
            reach,
            node=Reachability.namespace,
            **pubsubkwargs
        )




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.test.mocksocket import TestSocket, TestTransport
from slixmpp.test.livesocket import TestLiveSocket
from slixmpp.test.slixtest import *




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
import slixmpp
from slixmpp.xmlstream import ElementBase, ET


class ChatState(ElementBase):

    """
    Example chat state stanzas:

    ::

        <message>
          <active xmlns="http://jabber.org/protocol/chatstates" />
        </message>

        <message>
          <paused xmlns="http://jabber.org/protocol/chatstates" />
        </message>

    """

    name = ''
    namespace = 'http://jabber.org/protocol/chatstates'
    plugin_attrib = 'chat_state'
    interfaces = {'chat_state'}
    sub_interfaces = interfaces
    is_extension = True

    states = {'active', 'composing', 'gone', 'inactive', 'paused'}

    def setup(self, xml=None):
        self.xml = ET.Element('')
        return True

    def get_chat_state(self):
        parent = self.parent()
        for state in self.states:
            state_xml = parent.xml.find('{%s}%s' % (self.namespace, state))
            if state_xml is not None:
                self.xml = state_xml
                return state
        return ''

    def set_chat_state(self, state):
        self.del_chat_state()
        parent = self.parent()
        if state in self.states:
            self.xml = ET.Element('{%s}%s' % (self.namespace, state))
            parent.append(self.xml)
        elif state not in [None, '']:
            raise ValueError('Invalid chat state')

    def del_chat_state(self):
        parent = self.parent()
        for state in self.states:
            state_xml = parent.xml.find('{%s}%s' % (self.namespace, state))
            if state_xml is not None:
                self.xml = ET.Element('')
                parent.xml.remove(state_xml)


class Active(ChatState):
    name = 'active'


class Composing(ChatState):
    name = 'composing'


class Gone(ChatState):
    name = 'gone'


class Inactive(ChatState):
    name = 'inactive'


class Paused(ChatState):
    name = 'paused'




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0066 import stanza
from slixmpp.plugins.xep_0066.stanza import OOB, OOBTransfer
from slixmpp.plugins.xep_0066.oob import XEP_0066


register_plugin(XEP_0066)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 nicoco
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from typing import Literal, Optional, cast

from slixmpp import register_stanza_plugin
from slixmpp.plugins.xep_0402.stanza import Extensions
from slixmpp.types import ClientTypes
from slixmpp.xmlstream import ElementBase

NS = "urn:xmpp:notification-settings:0"

WhenLiteral = Literal["never", "always", "on-mention"]


class Notify(ElementBase):
    """
    Chat notification settings element


    To enable it on a Conference element, use configure() like this:

    .. code-block::python

        # C being a Conference element
        C['extensions']["notify"].configure("always", client_type="pc")

    Which will add the <notify> element to the <extensions> element.
    """

    namespace = NS
    name = "notify"
    plugin_attrib = "notify"
    interfaces = {"notify"}

    def configure(self, when: WhenLiteral, client_type: Optional[ClientTypes] = None) -> None:
        """
        Configure the chat notification settings for this bookmark.

        This method ensures that there are no conflicting settings, e.g.,
        both a <never /> and a <always /> element.
        """
        cls = _CLASS_MAP[when]
        element = cls()
        if client_type is not None:
            element["client-type"] = client_type

        match = client_type if client_type is not None else ""
        for child in self:
            if isinstance(child, _Base) and child["client-type"] == match:
                self.xml.remove(child.xml)

        self.append(element)

    def get_config(
        self, client_type: Optional[ClientTypes] = None
    ) -> Optional[WhenLiteral]:
        """
        Get the chat notification settings for this bookmark.

        :param client_type: Optionally, get the notification for a specific client type.
            If unset, returns the global notification setting.

        :return: The chat notification setting as a string, or None if unset.
        """
        match = client_type if client_type is not None else ""
        for child in self:
            if isinstance(child, _Base) and child["client-type"] == match:
                return cast(WhenLiteral, child.name)
        return None


class _Base(ElementBase):
    namespace = NS
    interfaces = {"client-type"}


class Never(_Base):
    name = "never"
    plugin_attrib = name


class Always(_Base):
    name = "always"
    plugin_attrib = name


class OnMention(_Base):
    name = "on-mention"
    plugin_attrib = name


class Advanced(ElementBase):
    namespace = NS
    name = plugin_attrib = "advanced"


_CLASS_MAP = {
    "never": Never,
    "always": Always,
    "on-mention": OnMention,
}


def register_plugin():
    register_stanza_plugin(Extensions, Notify)
    register_stanza_plugin(Notify, Advanced)
    register_stanza_plugin(Notify, Never, iterable=True)
    register_stanza_plugin(Notify, Always, iterable=True)
    register_stanza_plugin(Notify, OnMention, iterable=True)



import unittest
from slixmpp.test.integration import SlixIntegration


class TestLastActivity(SlixIntegration):
    async def asyncSetUp(self):
        await super().asyncSetUp()
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        self.register_plugins(['xep_0012'])
        await self.connect_clients()

    async def test_activity(self):
        """Check we can set and get last activity"""
        await self.clients[0]['xep_0012'].set_last_activity(
            status='coucou',
            seconds=4242,
        )
        act = await self.clients[1]['xep_0012'].get_last_activity(
            self.clients[0].boundjid.full
        )
        self.assertEqual(act['last_activity']['status'], 'coucou')
        self.assertGreater(act['last_activity']['seconds'], 4241)
        self.assertGreater(4250, act['last_activity']['seconds'])


suite = unittest.TestLoader().loadTestsFromTestCase(TestLastActivity)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.roster.item import RosterItem
from slixmpp.roster.single import RosterNode
from slixmpp.roster.multi import Roster




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0095 import stanza
from slixmpp.plugins.xep_0095.stanza import SI
from slixmpp.plugins.xep_0095.stream_initiation import XEP_0095


register_plugin(XEP_0095)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.stanza import Message
from slixmpp.xmlstream import (
    ElementBase,
    register_stanza_plugin,
)


NS = 'urn:xmpp:fasten:0'


class ApplyTo(ElementBase):
    namespace = NS
    name = 'apply-to'
    plugin_attrib = 'apply_to'
    interfaces = {'id', 'shell'}

    def set_shell(self, value: bool):
        if value:
            self.xml.attrib['shell'] = str(value).lower()
        else:
            if 'shell' in self.xml.attrib:
                del self.xml.attrib['shell']


class External(ElementBase):
    namespace = NS
    name = 'external'
    plugin_attrib = 'external'
    interfaces = {'name'}


def register_plugins():
    register_stanza_plugin(Message, ApplyTo)
    register_stanza_plugin(ApplyTo, External)
